# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Implements objects to represent problems, theorems, proofs, traceback."""

from __future__ import annotations

from collections import defaultdict
from typing import Any

import newclid.geometry as gm
import newclid.pretty as pt


def reshape(list_to_reshape: list[Any], n: int = 1) -> list[list[Any]]:
    assert len(list_to_reshape) % n == 0
    columns = [[] for i in range(n)]
    for i, x in enumerate(list_to_reshape):
        columns[i % n].append(x)
    return zip(*columns)


def isint(x: str) -> bool:
    try:
        int(x)
        return True
    except ValueError:
        return False


class Construction:
    """One predicate."""

    @classmethod
    def from_txt(cls, data: str) -> Construction:
        data = data.split(" ")
        return Construction(data[0], data[1:])

    def __init__(self, name: str, args: list[str]):
        self.name = name
        self.args = args

    def translate(self, mapping: dict[str, str]) -> Construction:
        args = [a if isint(a) else mapping[a] for a in self.args]
        return Construction(self.name, args)

    def txt(self) -> str:
        return " ".join([self.name] + list(self.args))


class Clause:
    """One construction (>= 1 predicate)."""

    @classmethod
    def from_txt(cls, data: str) -> Clause:
        if data == " =":
            return Clause([], [])
        points, constructions = data.split(" = ")
        return Clause(
            points.split(" "),
            [Construction.from_txt(c) for c in constructions.split(", ")],
        )

    def __init__(self, points: list[str], constructions: list[Construction]):
        self.points = []
        self.nums = []

        for p in points:
            num = None
            if isinstance(p, str) and "@" in p:
                p, num = p.split("@")
                x, y = num.split("_")
                num = float(x), float(y)
            self.points.append(p)
            self.nums.append(num)

        self.constructions = constructions

    def translate(self, mapping: dict[str, str]) -> Clause:
        points0 = []
        for p in self.points:
            pcount = len(mapping) + 1
            name = chr(96 + pcount)
            if name > "z":  # pcount = 26 -> name = 'z'
                name = chr(97 + (pcount - 1) % 26) + str((pcount - 1) // 26)

            p0 = mapping.get(p, name)
            mapping[p] = p0
            points0.append(p0)
        return Clause(points0, [c.translate(mapping) for c in self.constructions])

    def add(self, name: str, args: list[str]) -> None:
        self.constructions.append(Construction(name, args))

    def txt(self) -> str:
        return (
            " ".join(self.points)
            + " = "
            + ", ".join(c.txt() for c in self.constructions)
        )


def _gcd(x: int, y: int) -> int:
    while y:
        x, y = y, x % y
    return x


def simplify(n: int, d: int) -> tuple[int, int]:
    g = _gcd(n, d)
    return (n // g, d // g)


def compare_fn(dep: Dependency) -> tuple[Dependency, str]:
    return (dep, pt.pretty(dep))


def sort_deps(deps: list[Dependency]) -> list[Dependency]:
    return sorted(deps, key=compare_fn)


class Problem:
    """Describe one problem to solve."""

    @classmethod
    def from_txt_file(cls, fname: str, to_dict: bool = False, translate: bool = True):
        """Load a problem from a text file."""
        with open(fname, "r") as f:
            lines = f.read().split("\n")

        lines = [line for line in lines if line]
        data = [
            cls.from_txt(url + "\n" + problem, translate)
            for (url, problem) in reshape(lines, 2)
        ]
        if to_dict:
            return cls.to_dict(data)
        return data

    @classmethod
    def from_txt(cls, data: str, translate: bool = True) -> Problem:
        """Load a problem from a str object."""
        url = ""
        if "\n" in data:
            url, data = data.split("\n")

        if " ? " in data:
            clauses, goal = data.split(" ? ")
            goal = Construction.from_txt(goal)
        else:
            clauses, goal = data, None

        clauses = clauses.split("; ")
        problem = Problem(
            url=url, clauses=[Clause.from_txt(c) for c in clauses], goal=goal
        )
        if translate:
            return problem.translate()
        return problem

    @classmethod
    def to_dict(cls, data: list[Problem]) -> dict[str, Problem]:
        return {p.url: p for p in data}

    def __init__(self, url: str, clauses: list[Clause], goal: Construction):
        self.url = url
        self.clauses = clauses
        self.goal = goal

    def copy(self) -> Problem:
        return Problem(self.url, list(self.clauses), self.goal)

    def translate(self) -> Problem:  # to single-char point names
        """Translate point names into alphabetical."""
        mapping = {}
        clauses = []

        for clause in self.clauses:
            clauses.append(clause.translate(mapping))

        if self.goal:
            goal = self.goal.translate(mapping)
        else:
            goal = self.goal

        p = Problem(self.url, clauses, goal)
        p.mapping = mapping
        return p

    def txt(self) -> str:
        return (
            "; ".join([c.txt() for c in self.clauses]) + " ? " + self.goal.txt()
            if self.goal
            else ""
        )

    def setup_str_from_problem(self, definitions: list[Definition]) -> str:
        """Construct the <theorem_premises> string from Problem object."""
        ref = 0

        string = []
        for clause in self.clauses:
            group = {}
            p2deps = defaultdict(list)
            for c in clause.constructions:
                cdef = definitions[c.name]

                if len(c.args) != len(cdef.construction.args):
                    assert len(c.args) + len(clause.points) == len(
                        cdef.construction.args
                    )
                    c.args = clause.points + c.args

                mapping = dict(zip(cdef.construction.args, c.args))
                for points, bs in cdef.basics:
                    points = tuple([mapping[x] for x in points])
                    for p in points:
                        group[p] = points

                    for b in bs:
                        args = [mapping[a] for a in b.args]
                        name = b.name
                        if b.name in ["s_angle", "aconst"]:
                            x, y, z, v = args
                            name = "aconst"
                            v = int(v)

                            if v < 0:
                                v = -v
                                x, z = z, x

                            m, n = simplify(int(v), 180)
                            args = [y, z, y, x, f"{m}pi/{n}"]

                        p2deps[points].append(hashed_txt(name, args))

            for k, v in p2deps.items():
                p2deps[k] = sort_deps(v)

            points = clause.points
            while points:
                p = points[0]
                gr = group[p]
                points = [x for x in points if x not in gr]

                deps_str = []
                for dep in p2deps[gr]:
                    ref_str = "{:02}".format(ref)
                    dep_str = pt.pretty(dep)

                    if dep[0] == "aconst":
                        m, n = map(int, dep[-1].split("pi/"))
                        mn = f"{m}. pi / {n}."
                        dep_str = " ".join(dep_str.split()[:-1] + [mn])

                    deps_str.append(dep_str + " " + ref_str)
                    ref += 1

                string.append(" ".join(gr) + " : " + " ".join(deps_str))

        string = "{S} " + " ; ".join([s.strip() for s in string])
        goal = self.goal
        string += " ? " + pt.pretty([goal.name] + goal.args)
        return string


def parse_rely(s: str) -> dict[str, str]:
    result = {}
    if not s:
        return result
    s = [x.strip() for x in s.split(",")]
    for x in s:
        a, b = x.split(":")
        a, b = a.strip().split(), b.strip().split()
        result.update({m: b for m in a})
    return result


class Definition:
    """Definitions of construction statements."""

    @classmethod
    def from_txt_file(cls, fname: str, to_dict: bool = False) -> Definition:
        with open(fname, "r") as f:
            lines = f.read()
        return cls.from_string(lines, to_dict)

    @classmethod
    def from_string(cls, string: str, to_dict: bool = False) -> Definition:
        lines = string.split("\n")
        data = [cls.from_txt("\n".join(group)) for group in reshape(lines, 6)]
        if to_dict:
            return cls.to_dict(data)
        return data

    @classmethod
    def to_dict(cls, data: list[Definition]) -> dict[str, Definition]:
        return {d.construction.name: d for d in data}

    @classmethod
    def from_txt(cls, data: str) -> Definition:
        """Load definitions from a str object."""
        construction, rely, deps, basics, numerics, _ = data.split("\n")
        basics = [] if not basics else [b.strip() for b in basics.split(";")]

        levels = []
        for bs in basics:
            if ":" in bs:
                points, bs = bs.split(":")
                points = points.strip().split()
            else:
                points = []
            if bs.strip():
                bs = [Construction.from_txt(b.strip()) for b in bs.strip().split(",")]
            else:
                bs = []
            levels.append((points, bs))

        numerics = [] if not numerics else numerics.split(", ")

        return Definition(
            construction=Construction.from_txt(construction),
            rely=parse_rely(rely),
            deps=Clause.from_txt(deps),
            basics=levels,
            numerics=[Construction.from_txt(c) for c in numerics],
        )

    def __init__(
        self,
        construction: Construction,
        rely: dict[str, str],
        deps: Clause,
        basics: list[tuple[list[str], list[Construction]]],
        numerics: list[Construction],
    ):
        self.construction = construction
        self.rely = rely
        self.deps = deps
        self.basics = basics
        self.numerics = numerics

        args = set()
        for num in numerics:
            args.update(num.args)

        self.points = []
        self.args = []
        for p in self.construction.args:
            if p in args:
                self.args.append(p)
            else:
                self.points.append(p)


class Theorem:
    """Deduction rule."""

    @classmethod
    def from_txt_file(cls, fname: str, to_dict: bool = False) -> Theorem:
        with open(fname, "r") as f:
            theorems = f.read()
        return cls.from_string(theorems, to_dict)

    @classmethod
    def from_string(cls, string: str, to_dict: bool = False) -> Theorem:
        """Load deduction rule from a str object."""
        theorems = string.split("\n")
        theorems = [line for line in theorems if line and not line.startswith("#")]
        theorems = [cls.from_txt(line) for line in theorems]

        for i, th in enumerate(theorems):
            th.rule_name = "r{:02}".format(i)

        if to_dict:
            result = {}
            for t in theorems:
                if t.name in result:
                    t.name += "_"
                result[t.rule_name] = t

            return result

        return theorems

    @classmethod
    def from_txt(cls, data: str) -> Theorem:
        premises, conclusion = data.split(" => ")
        premises = premises.split(", ")
        conclusion = conclusion.split(", ")
        return Theorem(
            premise=[Construction.from_txt(p) for p in premises],
            conclusion=[Construction.from_txt(c) for c in conclusion],
        )

    def __init__(self, premise: list[Construction], conclusion: list[Construction]):
        if len(conclusion) != 1:
            raise ValueError("Cannot have more than one conclusion")
        self.name = "_".join([p.name for p in premise + conclusion])
        self.premise = premise
        self.conclusion = conclusion
        self.is_arg_reduce = False

        assert len(self.conclusion) == 1
        con = self.conclusion[0]

        if con.name in [
            "eqratio3",
            "midp",
            "contri",
            "simtri",
            "contri2",
            "simtri2",
            "simtri*",
            "contri*",
        ]:
            return

        prem_args = set(sum([p.args for p in self.premise], []))
        con_args = set(con.args)
        if len(prem_args) <= len(con_args):
            self.is_arg_reduce = True

    def txt(self) -> str:
        premise_txt = ", ".join([clause.txt() for clause in self.premise])
        conclusion_txt = ", ".join([clause.txt() for clause in self.conclusion])
        return f"{premise_txt} => {conclusion_txt}"

    def conclusion_name_args(
        self, mapping: dict[str, gm.Point]
    ) -> tuple[str, list[gm.Point]]:
        mapping = {arg: p for arg, p in mapping.items() if isinstance(arg, str)}
        c = self.conclusion[0]
        args = [mapping[a] for a in c.args]
        return c.name, args


def why_eqratio(
    d1: gm.Direction,
    d2: gm.Direction,
    d3: gm.Direction,
    d4: gm.Direction,
    level: int,
) -> list[Dependency]:
    """Why two ratios are equal, returns a Dependency objects."""
    all12 = list(gm.all_ratios(d1, d2, level))
    all34 = list(gm.all_ratios(d3, d4, level))

    min_why = None
    for ang12, d1s, d2s in all12:
        for ang34, d3s, d4s in all34:
            why0 = gm.why_equal(ang12, ang34, level)
            if why0 is None:
                continue
            d1_, d2_ = ang12._l
            d3_, d4_ = ang34._l
            why1 = gm.bfs_backtrack(d1, [d1_], d1s)
            why2 = gm.bfs_backtrack(d2, [d2_], d2s)
            why3 = gm.bfs_backtrack(d3, [d3_], d3s)
            why4 = gm.bfs_backtrack(d4, [d4_], d4s)
            why = why0 + why1 + why2 + why3 + why4
            if min_why is None or len(why) < len(min_why[0]):
                min_why = why, ang12, ang34, why0, why1, why2, why3, why4

    if min_why is None:
        return None

    _, ang12, ang34, why0, why1, why2, why3, why4 = min_why
    d1_, d2_ = ang12._l
    d3_, d4_ = ang34._l

    if d1 == d1_ and d2 == d2_ and d3 == d3_ and d4 == d4_:
        return why0

    (a_, b_), (c_, d_) = d1_._obj.points, d2_._obj.points
    (e_, f_), (g_, h_) = d3_._obj.points, d4_._obj.points
    deps = []
    if why0:
        dep = Dependency("eqratio", [a_, b_, c_, d_, e_, f_, g_, h_], "", level)
        dep.why = why0
        deps.append(dep)

    (a, b), (c, d) = d1._obj.points, d2._obj.points
    (e, f), (g, h) = d3._obj.points, d4._obj.points
    for why, (x, y), (x_, y_) in zip(
        [why1, why2, why3, why4],
        [(a, b), (c, d), (e, f), (g, h)],
        [(a_, b_), (c_, d_), (e_, f_), (g_, h_)],
    ):
        if why:
            dep = Dependency("cong", [x, y, x_, y_], "", level)
            dep.why = why
            deps.append(dep)

    return deps


def why_eqangle(
    d1: gm.Direction,
    d2: gm.Direction,
    d3: gm.Direction,
    d4: gm.Direction,
    level: int,
    verbose: bool = False,
) -> list[Dependency]:
    """Why two angles are equal, returns a Dependency objects."""
    all12 = list(gm.all_angles(d1, d2, level))
    all34 = list(gm.all_angles(d3, d4, level))

    min_why = None
    for ang12, d1s, d2s in all12:
        for ang34, d3s, d4s in all34:
            why0 = gm.why_equal(ang12, ang34, level)
            if why0 is None:
                continue
            d1_, d2_ = ang12._d
            d3_, d4_ = ang34._d
            why1 = gm.bfs_backtrack(d1, [d1_], d1s)
            why2 = gm.bfs_backtrack(d2, [d2_], d2s)
            why3 = gm.bfs_backtrack(d3, [d3_], d3s)
            why4 = gm.bfs_backtrack(d4, [d4_], d4s)
            why = why0 + why1 + why2 + why3 + why4
            if min_why is None or len(why) < len(min_why[0]):
                min_why = why, ang12, ang34, why0, why1, why2, why3, why4

    if min_why is None:
        return None

    _, ang12, ang34, why0, why1, why2, why3, why4 = min_why
    why0 = gm.why_equal(ang12, ang34, level)
    d1_, d2_ = ang12._d
    d3_, d4_ = ang34._d

    if d1 == d1_ and d2 == d2_ and d3 == d3_ and d4 == d4_:
        return (d1_, d2_, d3_, d4_), why0

    (a_, b_), (c_, d_) = d1_._obj.points, d2_._obj.points
    (e_, f_), (g_, h_) = d3_._obj.points, d4_._obj.points
    deps = []
    if why0:
        dep = Dependency("eqangle", [a_, b_, c_, d_, e_, f_, g_, h_], "", None)
        dep.why = why0
        deps.append(dep)

    (a, b), (c, d) = d1._obj.points, d2._obj.points
    (e, f), (g, h) = d3._obj.points, d4._obj.points
    for why, d_xy, (x, y), d_xy_, (x_, y_) in zip(
        [why1, why2, why3, why4],
        [d1, d2, d3, d4],
        [(a, b), (c, d), (e, f), (g, h)],
        [d1_, d2_, d3_, d4_],
        [(a_, b_), (c_, d_), (e_, f_), (g_, h_)],
    ):
        xy, xy_ = d_xy._obj, d_xy_._obj
        if why:
            if xy == xy_:
                name = "collx"
            else:
                name = "para"
            dep = Dependency(name, [x_, y_, x, y], "", None)
            dep.why = why
            deps.append(dep)

    return (d1_, d2_, d3_, d4_), deps


CONSTRUCTION_RULE = "c0"


class EmptyDependency:
    """Empty dependency predicate ready to get filled up."""

    def __init__(self, level: int, rule_name: str):
        self.level = level
        self.rule_name = rule_name or ""
        self.empty = True
        self.why = []
        self.trace = None

    def populate(self, name: str, args: list[gm.Point]) -> Dependency:
        dep = Dependency(name, args, self.rule_name, self.level)
        dep.trace2 = self.trace
        dep.why = list(self.why)
        return dep

    def copy(self) -> EmptyDependency:
        other = EmptyDependency(self.level, self.rule_name)
        other.why = list(self.why)
        return other

    def extend(
        self,
        g: Any,
        name0: str,
        args0: list[gm.Point],
        name: str,
        args: list[gm.Point],
    ) -> EmptyDependency:
        """Extend the dependency list by (name, args)."""
        dep0 = self.populate(name0, args0)
        deps = EmptyDependency(level=self.level, rule_name=None)
        dep = Dependency(name, args, None, deps.level)
        deps.why = [dep0, dep.why_me_or_cache(g, None)]
        return deps

    def extend_many(
        self,
        g: Any,
        name0: str,
        args0: list[gm.Point],
        name_args: list[tuple[str, list[gm.Point]]],
    ) -> EmptyDependency:
        """Extend the dependency list by many name_args."""
        if not name_args:
            return self
        dep0 = self.populate(name0, args0)
        deps = EmptyDependency(level=self.level, rule_name=None)
        deps.why = [dep0]
        for name, args in name_args:
            dep = Dependency(name, args, None, deps.level)
            deps.why += [dep.why_me_or_cache(g, None)]
        return deps


def maybe_make_equal_pairs(
    a: gm.Point,
    b: gm.Point,
    c: gm.Point,
    d: gm.Point,
    m: gm.Point,
    n: gm.Point,
    p: gm.Point,
    q: gm.Point,
    ab: gm.Line,
    mn: gm.Line,
    g: Any,
    level: int,
) -> list[Dependency]:
    """Make a-b:c-d==m-n:p-q in case a-b==m-n or c-d==p-q."""
    if ab != mn:
        return
    why = []
    eqname = "para" if isinstance(ab, gm.Line) else "cong"
    colls = [a, b, m, n]
    if len(set(colls)) > 2 and eqname == "para":
        dep = Dependency("collx", colls, None, level)
        dep.why_me(g, level)
        why += [dep]

    dep = Dependency(eqname, [c, d, p, q], None, level)
    dep.why_me(g, level)
    why += [dep]
    return why


class Dependency(Construction):
    """Dependency is a predicate that other predicates depend on."""

    def __init__(self, name: str, args: list[gm.Point], rule_name: str, level: int):
        super().__init__(name, args)
        self.rule_name = rule_name or ""
        self.level = level
        self.why = []

        self._stat = None
        self.trace = None

    def _find(self, dep_hashed: tuple[str, ...]) -> Dependency:
        for w in self.why:
            f = w._find(dep_hashed)
            if f:
                return f
            if w.hashed() == dep_hashed:
                return w

    def remove_loop(self) -> Dependency:
        f = self._find(self.hashed())
        if f:
            return f
        return self

    def copy(self) -> Dependency:
        dep = Dependency(self.name, self.args, self.rule_name, self.level)
        dep.trace = self.trace
        dep.why = list(self.why)
        return dep

    def why_me_or_cache(self, g: Any, level: int) -> Dependency:
        if self.hashed() in g.cache:
            return g.cache[self.hashed()]
        self.why_me(g, level)
        return self

    def populate(self, name: str, args: list[gm.Point]) -> Dependency:
        assert self.rule_name == CONSTRUCTION_RULE, self.rule_name
        dep = Dependency(self.name, self.args, self.rule_name, self.level)
        dep.why = list(self.why)
        return dep

    def why_me(self, g: Any, level: int) -> None:
        """Figure out the dependencies predicates of self."""
        name, args = self.name, self.args

        hashed_me = hashed(name, args)
        if hashed_me in g.cache:
            dep = g.cache[hashed_me]
            self.why = dep.why
            self.rule_name = dep.rule_name
            return

        if self.name == "para":
            a, b, c, d = self.args
            if {a, b} == {c, d}:
                self.why = []
                return

            ab = g._get_line(a, b)
            cd = g._get_line(c, d)
            if ab == cd:
                if {a, b} == {c, d}:
                    self.why = []
                    self.rule_name = ""
                    return
                dep = Dependency("coll", list({a, b, c, d}), "t??", None)
                self.why = [dep.why_me_or_cache(g, level)]
                return

            for (x, y), xy in zip([(a, b), (c, d)], [ab, cd]):
                x_, y_ = xy.points
                if {x, y} == {x_, y_}:
                    continue
                d = Dependency("collx", [x, y, x_, y_], None, level)
                self.why += [d.why_me_or_cache(g, level)]

            whypara = g.why_equal(ab, cd, None)
            self.why += whypara

        elif self.name == "midp":
            m, a, b = self.args
            ma = g._get_segment(m, a)
            mb = g._get_segment(m, b)
            dep = Dependency("coll", [m, a, b], None, None).why_me_or_cache(g, None)
            self.why = [dep] + g.why_equal(ma, mb, level)

        elif self.name == "perp":
            a, b, c, d = self.args
            ab = g._get_line(a, b)
            cd = g._get_line(c, d)
            for (x, y), xy in zip([(a, b), (c, d)], [ab, cd]):
                x_, y_ = xy.points
                if {x, y} == {x_, y_}:
                    continue
                d = Dependency("collx", [x, y, x_, y_], None, level)
                self.why += [d.why_me_or_cache(g, level)]

            _, why = why_eqangle(ab._val, cd._val, cd._val, ab._val, level)
            a, b = ab.points
            c, d = cd.points

            if hashed(self.name, [a, b, c, d]) != self.hashed():
                d = Dependency(self.name, [a, b, c, d], None, level)
                d.why = why
                why = [d]

            self.why += why

        elif self.name == "cong":
            a, b, c, d = self.args
            ab = g._get_segment(a, b)
            cd = g._get_segment(c, d)

            self.why = g.why_equal(ab, cd, level)

        elif self.name == "coll":
            _, why = gm.line_of_and_why(self.args, level)
            self.why = why

        elif self.name == "collx":
            if g.check_coll(self.args):
                args = list(set(self.args))
                hashed_me = hashed("coll", args)
                if hashed_me in g.cache:
                    dep = g.cache[hashed_me]
                    self.why = [dep]
                    self.rule_name = ""
                    return
                _, self.why = gm.line_of_and_why(args, level)
            else:
                self.name = "para"
                self.why_me(g, level)

        elif self.name == "cyclic":
            _, why = gm.circle_of_and_why(self.args, level)
            self.why = why

        elif self.name == "circle":
            o, a, b, c = self.args
            oa = g._get_segment(o, a)
            ob = g._get_segment(o, b)
            oc = g._get_segment(o, c)
            self.why = g.why_equal(oa, ob, level) + g.why_equal(oa, oc, level)

        elif self.name in ["eqangle", "eqangle6"]:
            a, b, c, d, m, n, p, q = self.args

            ab, why1 = g.get_line_thru_pair_why(a, b)
            cd, why2 = g.get_line_thru_pair_why(c, d)
            mn, why3 = g.get_line_thru_pair_why(m, n)
            pq, why4 = g.get_line_thru_pair_why(p, q)

            if ab is None or cd is None or mn is None or pq is None:
                if {a, b} == {m, n}:
                    d = Dependency("para", [c, d, p, q], None, level)
                    self.why = [d.why_me_or_cache(g, level)]
                if {a, b} == {c, d}:
                    d = Dependency("para", [p, q, m, n], None, level)
                    self.why = [d.why_me_or_cache(g, level)]
                if {c, d} == {p, q}:
                    d = Dependency("para", [a, b, m, n], None, level)
                    self.why = [d.why_me_or_cache(g, level)]
                if {p, q} == {m, n}:
                    d = Dependency("para", [a, b, c, d], None, level)
                    self.why = [d.why_me_or_cache(g, level)]
                return

            for (x, y), xy, whyxy in zip(
                [(a, b), (c, d), (m, n), (p, q)],
                [ab, cd, mn, pq],
                [why1, why2, why3, why4],
            ):
                x_, y_ = xy.points
                if {x, y} == {x_, y_}:
                    continue
                d = Dependency("collx", [x, y, x_, y_], None, level)
                d.why = whyxy
                self.why += [d]

            a, b = ab.points
            c, d = cd.points
            m, n = mn.points
            p, q = pq.points
            diff = hashed(self.name, [a, b, c, d, m, n, p, q]) != self.hashed()

            whyeqangle = None
            if ab._val and cd._val and mn._val and pq._val:
                whyeqangle = why_eqangle(ab._val, cd._val, mn._val, pq._val, level)

            if whyeqangle:
                (dab, dcd, dmn, dpq), whyeqangle = whyeqangle
                if diff:
                    d = Dependency("eqangle", [a, b, c, d, m, n, p, q], None, level)
                    d.why = whyeqangle
                    whyeqangle = [d]
                self.why += whyeqangle

            else:
                if (ab == cd and mn == pq) or (ab == mn and cd == pq):
                    self.why += []
                elif ab == mn:
                    self.why += maybe_make_equal_pairs(
                        a, b, c, d, m, n, p, q, ab, mn, g, level
                    )
                elif cd == pq:
                    self.why += maybe_make_equal_pairs(
                        c, d, a, b, p, q, m, n, cd, pq, g, level
                    )
                elif ab == cd:
                    self.why += maybe_make_equal_pairs(
                        a, b, m, n, c, d, p, q, ab, cd, g, level
                    )
                elif mn == pq:
                    self.why += maybe_make_equal_pairs(
                        m, n, a, b, p, q, c, d, mn, pq, g, level
                    )
                elif g.is_equal(ab, mn) or g.is_equal(cd, pq):
                    dep1 = Dependency("para", [a, b, m, n], None, level)
                    dep1.why_me(g, level)
                    dep2 = Dependency("para", [c, d, p, q], None, level)
                    dep2.why_me(g, level)
                    self.why += [dep1, dep2]
                elif g.is_equal(ab, cd) or g.is_equal(mn, pq):
                    dep1 = Dependency("para", [a, b, c, d], None, level)
                    dep1.why_me(g, level)
                    dep2 = Dependency("para", [m, n, p, q], None, level)
                    dep2.why_me(g, level)
                    self.why += [dep1, dep2]
                elif ab._val and cd._val and mn._val and pq._val:
                    self.why = why_eqangle(ab._val, cd._val, mn._val, pq._val, level)

        elif self.name in ["eqratio", "eqratio6"]:
            a, b, c, d, m, n, p, q = self.args
            ab = g._get_segment(a, b)
            cd = g._get_segment(c, d)
            mn = g._get_segment(m, n)
            pq = g._get_segment(p, q)

            if ab is None or cd is None or mn is None or pq is None:
                if {a, b} == {m, n}:
                    d = Dependency("cong", [c, d, p, q], None, level)
                    self.why = [d.why_me_or_cache(g, level)]
                if {a, b} == {c, d}:
                    d = Dependency("cong", [p, q, m, n], None, level)
                    self.why = [d.why_me_or_cache(g, level)]
                if {c, d} == {p, q}:
                    d = Dependency("cong", [a, b, m, n], None, level)
                    self.why = [d.why_me_or_cache(g, level)]
                if {p, q} == {m, n}:
                    d = Dependency("cong", [a, b, c, d], None, level)
                    self.why = [d.why_me_or_cache(g, level)]
                return

            if ab._val and cd._val and mn._val and pq._val:
                self.why = why_eqratio(ab._val, cd._val, mn._val, pq._val, level)

            if self.why is None:
                self.why = []
                if (ab == cd and mn == pq) or (ab == mn and cd == pq):
                    self.why = []
                elif ab == mn:
                    self.why += maybe_make_equal_pairs(
                        a, b, c, d, m, n, p, q, ab, mn, g, level
                    )
                elif cd == pq:
                    self.why += maybe_make_equal_pairs(
                        c, d, a, b, p, q, m, n, cd, pq, g, level
                    )
                elif ab == cd:
                    self.why += maybe_make_equal_pairs(
                        a, b, m, n, c, d, p, q, ab, cd, g, level
                    )
                elif mn == pq:
                    self.why += maybe_make_equal_pairs(
                        m, n, a, b, p, q, c, d, mn, pq, g, level
                    )
                elif g.is_equal(ab, mn) or g.is_equal(cd, pq):
                    dep1 = Dependency("cong", [a, b, m, n], None, level)
                    dep1.why_me(g, level)
                    dep2 = Dependency("cong", [c, d, p, q], None, level)
                    dep2.why_me(g, level)
                    self.why += [dep1, dep2]
                elif g.is_equal(ab, cd) or g.is_equal(mn, pq):
                    dep1 = Dependency("cong", [a, b, c, d], None, level)
                    dep1.why_me(g, level)
                    dep2 = Dependency("cong", [m, n, p, q], None, level)
                    dep2.why_me(g, level)
                    self.why += [dep1, dep2]
                elif ab._val and cd._val and mn._val and pq._val:
                    self.why = why_eqangle(ab._val, cd._val, mn._val, pq._val, level)

        elif self.name in ["diff", "npara", "nperp", "ncoll", "sameside"]:
            self.why = []

        elif self.name == "simtri":
            a, b, c, x, y, z = self.args
            dep1 = Dependency("eqangle", [a, b, a, c, x, y, x, z], "", level)
            dep1.why_me(g, level)
            dep2 = Dependency("eqangle", [b, a, b, c, y, x, y, z], "", level)
            dep2.why_me(g, level)
            self.rule_name = "r34"
            self.why = [dep1, dep2]

        elif self.name == "contri":
            a, b, c, x, y, z = self.args
            dep1 = Dependency("cong", [a, b, x, y], "", level)
            dep1.why_me(g, level)
            dep2 = Dependency("cong", [b, c, y, z], "", level)
            dep2.why_me(g, level)
            dep3 = Dependency("cong", [c, a, z, x], "", level)
            dep3.why_me(g, level)
            self.rule_name = "r32"
            self.why = [dep1, dep2, dep3]

        elif self.name == "ind":
            pass

        elif self.name == "aconst":
            a, b, c, d, ang0 = self.args

            measure = ang0._val

            for ang in measure.neighbors(gm.Angle):
                if ang == ang0:
                    continue
                d1, d2 = ang._d
                l1, l2 = d1._obj, d2._obj
                (a1, b1), (c1, d1) = l1.points, l2.points

                if not g.check_para_or_coll([a, b, a1, b1]) or not g.check_para_or_coll(
                    [c, d, c1, d1]
                ):
                    continue

                self.why = []
                for args in [(a, b, a1, b1), (c, d, c1, d1)]:
                    if g.check_coll(args):
                        if len(set(args)) > 2:
                            dep = Dependency("coll", args, None, None)
                            self.why.append(dep.why_me_or_cache(g, level))
                    else:
                        dep = Dependency("para", args, None, None)
                        self.why.append(dep.why_me_or_cache(g, level))

                self.why += gm.why_equal(ang, ang0)
                break

        elif self.name == "rconst":
            a, b, c, d, rat0 = self.args

            val = rat0._val

            for rat in val.neighbors(gm.Ratio):
                if rat == rat0:
                    continue
                l1, l2 = rat._l
                s1, s2 = l1._obj, l2._obj
                (a1, b1), (c1, d1) = list(s1.points), list(s2.points)

                if not g.check_cong([a, b, a1, b1]) or not g.check_cong([c, d, c1, d1]):
                    continue

                self.why = []
                for args in [(a, b, a1, b1), (c, d, c1, d1)]:
                    if len(set(args)) > 2:
                        dep = Dependency("cong", args, None, None)
                        self.why.append(dep.why_me_or_cache(g, level))

                self.why += gm.why_equal(rat, rat0)
                break

        else:
            raise ValueError("Not recognize", self.name)

    def hashed(self, rename: bool = False) -> tuple[str, ...]:
        return hashed(self.name, self.args, rename=rename)


def hashed(name: str, args: list[gm.Point], rename: bool = False) -> tuple[str, ...]:
    if name == "s_angle":
        args = [p.name if not rename else p.new_name for p in args[:-1]] + [
            str(args[-1])
        ]
    else:
        args = [p.name if not rename else p.new_name for p in args]
    return hashed_txt(name, args)


def hashed_txt(name: str, args: list[str]) -> tuple[str, ...]:
    """Return a tuple unique to name and args upto arg permutation equivariant."""

    if name in ["const", "aconst", "rconst"]:
        a, b, c, d, y = args
        a, b = sorted([a, b])
        c, d = sorted([c, d])
        return name, a, b, c, d, y

    if name in ["npara", "nperp", "para", "cong", "perp", "collx"]:
        a, b, c, d = args

        a, b = sorted([a, b])
        c, d = sorted([c, d])
        (a, b), (c, d) = sorted([(a, b), (c, d)])

        return (name, a, b, c, d)

    if name in ["midp", "midpoint"]:
        a, b, c = args
        b, c = sorted([b, c])
        return (name, a, b, c)

    if name in ["coll", "cyclic", "ncoll", "diff", "triangle"]:
        return (name,) + tuple(sorted(list(set(args))))

    if name == "circle":
        x, a, b, c = args
        return (name, x) + tuple(sorted([a, b, c]))

    if name in ["eqangle", "eqratio", "eqangle6", "eqratio6"]:
        a, b, c, d, e, f, g, h = args
        a, b = sorted([a, b])
        c, d = sorted([c, d])
        e, f = sorted([e, f])
        g, h = sorted([g, h])
        if tuple(sorted([a, b, e, f])) > tuple(sorted([c, d, g, h])):
            a, b, e, f, c, d, g, h = c, d, g, h, a, b, e, f
        if (a, b, c, d) > (e, f, g, h):
            a, b, c, d, e, f, g, h = e, f, g, h, a, b, c, d

        if name == "eqangle6":
            name = "eqangle"
        if name == "eqratio6":
            name = "eqratio"
        return (name,) + (a, b, c, d, e, f, g, h)

    if name in ["contri", "simtri", "simtri2", "contri2", "contri*", "simtri*"]:
        a, b, c, x, y, z = args
        (a, x), (b, y), (c, z) = sorted([(a, x), (b, y), (c, z)], key=sorted)
        (a, b, c), (x, y, z) = sorted([(a, b, c), (x, y, z)], key=sorted)
        return (name, a, b, c, x, y, z)

    if name in ["eqratio3"]:
        a, b, c, d, o, o = args
        (a, c), (b, d) = sorted([(a, c), (b, d)], key=sorted)
        (a, b), (c, d) = sorted([(a, b), (c, d)], key=sorted)
        return (name, a, b, c, d, o, o)

    if name in ["sameside", "s_angle"]:
        return (name,) + tuple(args)

    raise ValueError(f"Not recognize {name} to hash.")



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Unit tests for newclid.py."""

import pytest
import pytest_check as check

from newclid.api import GeometricSolverBuilder


class TestDDAR:
    @pytest.fixture(autouse=True)
    def setUpClass(self):
        self.solver_builder = GeometricSolverBuilder()

    def test_orthocenter_should_fail(self):
        solver = self.solver_builder.load_problem_from_txt(
            "a b c = triangle a b c; "
            "d = on_tline d b a c, on_tline d c a b "
            "? perp a d b c"
        ).build()
        success = solver.run()
        check.is_false(success)

    def test_orthocenter_aux_should_succeed(self):
        solver = self.solver_builder.load_problem_from_txt(
            "a b c = triangle a b c; "
            "d = on_tline d b a c, on_tline d c a b; "
            "e = on_line e a c, on_line e b d "
            "? perp a d b c"
        ).build()
        success = solver.run()
        check.is_true(success)

    def test_incenter_excenter_should_succeed(self):
        # Note that this same problem should fail in dd_test.py
        solver = self.solver_builder.load_problem_from_txt(
            "a b c = triangle a b c; "
            "d = incenter d a b c; "
            "e = excenter e a b c "
            "? perp d c c e"
        ).build()
        success = solver.run()
        check.is_true(success)



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Unit tests for ar.py."""

import pytest
import pytest_check as check
from newclid.api import GeometricSolverBuilder
import newclid.ar as ar
import newclid.problem as pr


class TestAR:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.solver_builder = GeometricSolverBuilder()

    def test_update_groups(self):
        """Test for update_groups."""
        groups1 = [{1, 2}, {3, 4, 5}, {6, 7}]
        groups2 = [{2, 3, 8}, {9, 10, 11}]

        _, links, history = ar.update_groups(groups1, groups2)
        check.equal(
            history,
            [
                [{1, 2, 3, 4, 5, 8}, {6, 7}],
                [{1, 2, 3, 4, 5, 8}, {6, 7}, {9, 10, 11}],
            ],
        )
        check.equal(links, [(2, 3), (3, 8), (9, 10), (10, 11)])

        groups1 = [{1, 2}, {3, 4}, {5, 6}, {7, 8}]
        groups2 = [{2, 3, 8, 9, 10}, {3, 6, 11}]

        _, links, history = ar.update_groups(groups1, groups2)
        check.equal(
            history,
            [
                [{1, 2, 3, 4, 7, 8, 9, 10}, {5, 6}],
                [{1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11}],
            ],
        )
        check.equal(links, [(2, 3), (3, 8), (8, 9), (9, 10), (3, 6), (6, 11)])

        groups1 = []
        groups2 = [{1, 2}, {3, 4}, {5, 6}, {2, 3}]

        _, links, history = ar.update_groups(groups1, groups2)
        check.equal(
            history,
            [
                [{1, 2}],
                [{1, 2}, {3, 4}],
                [{1, 2}, {3, 4}, {5, 6}],
                [{1, 2, 3, 4}, {5, 6}],
            ],
        )
        check.equal(links, [(1, 2), (3, 4), (5, 6), (2, 3)])

    def test_generic_table_simple(self):
        tb = ar.Table()

        # If a-b = b-c & d-a = c-d
        tb.add_eq4("a", "b", "b", "c", "fact1")
        tb.add_eq4("d", "a", "c", "d", "fact2")
        tb.add_eq4("x", "y", "z", "t", "fact3")  # distractor fact

        # Then b=d, because {fact1, fact2} but not fact3.
        result = list(tb.get_all_eqs_and_why())
        check.is_in(("b", "d", ["fact1", "fact2"]), result)

    def test_angle_table_inbisector_exbisector(self):
        """Test that AR can figure out bisector & ex-bisector are perpendicular."""
        # Load the scenario that we have cd is bisector of acb and
        # ce is the ex-bisector of acb.

        solver = self.solver_builder.load_problem_from_txt(
            "a b c = triangle a b c; "
            "d = incenter d a b c; "
            "e = excenter e a b c "
            "? perp d c c e"
        ).build()
        g = solver.proof_state

        # Create an external angle table:
        tb = ar.AngleTable("pi")

        # Add bisector & ex-bisector facts into the table:
        ca, cd, cb, ce = g.names2nodes(["d(ac)", "d(cd)", "d(bc)", "d(ce)"])
        tb.add_eqangle(ca, cd, cd, cb, "fact1")
        tb.add_eqangle(ce, ca, cb, ce, "fact2")

        # Add a distractor fact to make sure traceback does not include this fact
        ab = g.names2nodes(["d(ab)"])[0]
        tb.add_eqangle(ab, cb, cb, ca, "fact3")

        # Check for all new equalities
        result = list(tb.get_all_eqs_and_why())

        # halfpi is represented as a tuple (1, 2)
        halfpi = (1, 2)

        # check that cd-ce == halfpi and this is because fact1 & fact2, not fact3
        check.equal(
            result,
            [
                (cd, ce, halfpi, ["fact1", "fact2"]),
                (ce, cd, halfpi, ["fact1", "fact2"]),
            ],
        )

    def test_angle_table_equilateral_triangle(self):
        """Test that AR can figure out triangles with 3 equal angles => each is pi/3."""
        # Load an equaliteral scenario
        solver = self.solver_builder.load_problem_from_txt(
            "a b c = ieq_triangle " "? cong a b a c"
        ).build()
        g = solver.proof_state

        # Add two eqangles facts because ieq_triangle only add congruent sides
        a, b, c = g.names2nodes("abc")
        g.add_eqangle([a, b, b, c, b, c, c, a], pr.EmptyDependency(0, None))
        g.add_eqangle([b, c, c, a, c, a, a, b], pr.EmptyDependency(0, None))

        # Create an external angle table:
        tb = ar.AngleTable("pi")

        # Add the fact that there are three equal angles
        ab, bc, ca = g.names2nodes(["d(ab)", "d(bc)", "d(ac)"])
        tb.add_eqangle(ab, bc, bc, ca, "fact1")
        tb.add_eqangle(bc, ca, ca, ab, "fact2")

        # Now check for all new equalities
        result = list(tb.get_all_eqs_and_why())
        result = [(x.name, y.name, z, t) for x, y, z, t in result]

        # 1/3 pi is represented as a tuple angle_60
        angle_60 = (1, 3)
        angle_120 = (2, 3)

        # check that angles constants are created and figured out:
        check.equal(
            result,
            [
                ("d(bc)", "d(ac)", angle_120, ["fact1", "fact2"]),
                ("d(ab)", "d(bc)", angle_120, ["fact1", "fact2"]),
                ("d(ac)", "d(ab)", angle_120, ["fact1", "fact2"]),
                ("d(ac)", "d(bc)", angle_60, ["fact1", "fact2"]),
                ("d(bc)", "d(ab)", angle_60, ["fact1", "fact2"]),
                ("d(ab)", "d(ac)", angle_60, ["fact1", "fact2"]),
            ],
        )

    def test_incenter_excenter_touchpoints(self):
        """Test that AR can figure out incenter/excenter touchpoints are equidistant to midpoint."""
        solver = self.solver_builder.load_problem_from_txt(
            "a b c = triangle a b c; "
            "d1 d2 d3 d = incenter2 a b c; "
            "e1 e2 e3 e = excenter2 a b c "
            "? perp d c c e",
            translate=False,
        ).build()
        g = solver.proof_state

        a, b, c, ab, bc, ca, d1, d2, d3, e1, e2, e3 = g.names2nodes(
            ["a", "b", "c", "ab", "bc", "ac", "d1", "d2", "d3", "e1", "e2", "e3"]
        )

        # Create an external distance table:
        tb = ar.DistanceTable()

        # DD can figure out the following facts,
        # we manually add them to AR.
        tb.add_cong(ab, ca, a, d3, a, d2, "fact1")
        tb.add_cong(ab, ca, a, e3, a, e2, "fact2")
        tb.add_cong(ca, bc, c, d2, c, d1, "fact5")
        tb.add_cong(ca, bc, c, e2, c, e1, "fact6")
        tb.add_cong(bc, ab, b, d1, b, d3, "fact3")
        tb.add_cong(bc, ab, b, e1, b, e3, "fact4")

        # Now we check whether tb has figured out that
        # distance(b, d1) == distance(e1, c)

        # linear comb exprssion of each variables:
        b = tb.v2e["bc:b"]
        c = tb.v2e["bc:c"]
        d1 = tb.v2e["bc:d1"]
        e1 = tb.v2e["bc:e1"]

        check.equal(ar.minus(d1, b), ar.minus(c, e1))



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Unit testing for the geometry numericals code."""

import pytest_check as check

import newclid.numericals as nm

np = nm.np

unif = nm.unif
Point = nm.Point
Line = nm.Line
Circle = nm.Circle
HalfLine = nm.HalfLine

line_circle_intersection = nm.line_circle_intersection
line_line_intersection = nm.line_line_intersection

check_coll = nm.check_coll
check_eqangle = nm.check_eqangle

random_points = nm.random_points
ang_between = nm.ang_between
head_from = nm.head_from


class TestNumerical:
    def test_sketch_ieq_triangle(self):
        a, b, c = nm.sketch_ieq_triangle([])
        check.almost_equal(a.distance(b), b.distance(c))
        check.almost_equal(c.distance(a), b.distance(c))

    def test_sketch_2l1c(self):
        p = nm.Point(0.0, 0.0)
        pi = np.pi
        anga = unif(-0.4 * pi, 0.4 * pi)
        a = Point(np.cos(anga), np.sin(anga))
        angb = unif(0.6 * pi, 1.4 * pi)
        b = Point(np.cos(angb), np.sin(angb))

        angc = unif(anga + 0.05 * pi, angb - 0.05 * pi)
        c = Point(np.cos(angc), np.sin(angc)) * unif(0.2, 0.8)

        x, y, z, i = nm.sketch_2l1c([a, b, c, p])
        check.is_true(check_coll([x, c, a]))
        check.is_true(check_coll([y, c, b]))
        check.almost_equal(z.distance(p), 1.0)
        check.is_true(check_coll([p, i, z]))
        check.is_true(Line(i, x).is_perp(Line(c, a)))
        check.is_true(Line(i, y).is_perp(Line(c, b)))
        check.almost_equal(i.distance(x), i.distance(y))
        check.almost_equal(i.distance(x), i.distance(z))

    def test_sketch_3peq(self):
        a, b, c = random_points(3)
        x, y, z = nm.sketch_3peq([a, b, c])

        check.is_true(check_coll([a, b, x]))
        check.is_true(check_coll([a, c, y]))
        check.is_true(check_coll([b, c, z]))
        check.is_true(check_coll([x, y, z]))
        check.almost_equal(z.distance(x), z.distance(y))

    def test_sketch_aline(self):
        a, b, c, d, e = random_points(5)
        ex = nm.sketch_aline([a, b, c, d, e])
        check.is_instance(ex, HalfLine)
        check.equal(ex.tail, e)
        x = ex.head
        check.almost_equal(ang_between(b, a, c), ang_between(e, d, x))

    def test_sketch_amirror(self):
        a, b, c = random_points(3)
        bx = nm.sketch_amirror([a, b, c])
        check.is_instance(bx, HalfLine)
        assert bx.tail == b
        x = bx.head

        ang1 = ang_between(b, a, c)
        ang2 = ang_between(b, c, x)
        check.almost_equal(ang1, ang2)

    def test_sketch_bisect(self):
        a, b, c = random_points(3)
        line = nm.sketch_bisect([a, b, c])
        check.almost_equal(b.distance(line), 0.0)

        perpendicular_line = a.perpendicular_line(line)
        x = line_line_intersection(perpendicular_line, Line(b, c))
        check.almost_equal(a.distance(line), x.distance(line))

        d, _ = line_circle_intersection(line, Circle(b, radius=1))
        ang1 = ang_between(b, a, d)
        ang2 = ang_between(b, d, c)
        check.almost_equal(ang1, ang2)

    def test_sketch_bline(self):
        a, b = random_points(2)
        line_ab = nm.sketch_bline([a, b])
        check.is_true(Line(a, b).is_perp(line_ab))
        check.almost_equal(a.distance(line_ab), b.distance(line_ab))

    def test_sketch_cc_tangent(self):
        o = Point(0.0, 0.0)
        w = Point(1.0, 0.0)

        ra = unif(0.0, 0.6)
        rb = unif(0.4, 1.0)

        a = unif(0.0, np.pi)
        b = unif(0.0, np.pi)

        a = o + ra * Point(np.cos(a), np.sin(a))
        b = w + rb * Point(np.sin(b), np.cos(b))

        x, y, z, t = nm.sketch_cc_tangent([o, a, w, b])
        xy = Line(x, y)
        zt = Line(z, t)
        check.almost_equal(o.distance(xy), o.distance(a))
        check.almost_equal(o.distance(zt), o.distance(a))
        check.almost_equal(w.distance(xy), w.distance(b))
        check.almost_equal(w.distance(zt), w.distance(b))

    def test_sketch_circle(self):
        a, b, c = random_points(3)
        circle = nm.sketch_circle([a, b, c])
        check.almost_equal(circle.center.distance(a), 0.0)
        check.almost_equal(circle.radius, b.distance(c))

    def test_sketch_e5128(self):
        b = Point(0.0, 0.0)
        c = Point(0.0, 1.0)
        ang = unif(-np.pi / 2, 3 * np.pi / 2)
        d = head_from(c, ang, 1.0)
        a = Point(unif(0.5, 2.0), 0.0)

        e, g = nm.sketch_e5128([a, b, c, d])
        ang1 = ang_between(a, b, d)
        ang2 = ang_between(e, a, g)
        check.almost_equal(ang1, ang2)

    def test_sketch_eq_quadrangle(self):
        a, b, c, d = nm.sketch_eq_quadrangle([])
        check.almost_equal(a.distance(d), c.distance(b))
        ac = Line(a, c)
        assert ac.diff_side(b, d), (ac(b), ac(d))
        bd = Line(b, d)
        assert bd.diff_side(a, c), (bd(a), bd(c))

    def test_sketch_eq_trapezoid(self):
        a, b, c, d = nm.sketch_eq_trapezoid([])
        assert Line(a, b).is_parallel(Line(c, d))
        check.almost_equal(a.distance(d), b.distance(c))

    def test_sketch_eqangle3(self):
        points = random_points(5)
        x = nm.sketch_eqangle3(points).sample_within(points)[0]
        a, b, d, e, f = points
        check.is_true(check_eqangle([x, a, x, b, d, e, d, f]))

    def test_sketch_eqangle2(self):
        a, b, c = random_points(3)
        x = nm.sketch_eqangle2([a, b, c])
        ang1 = ang_between(a, b, x)
        ang2 = ang_between(c, x, b)
        check.almost_equal(ang1, ang2)

    def test_sketch_edia_quadrangle(self):
        a, b, c, d = nm.sketch_eqdia_quadrangle([])
        assert Line(a, c).diff_side(b, d)
        assert Line(b, d).diff_side(a, c)
        check.almost_equal(a.distance(c), b.distance(d))

    def test_sketch_isos(self):
        a, b, c = nm.sketch_isos([])
        check.almost_equal(a.distance(b), a.distance(c))
        check.almost_equal(ang_between(b, a, c), ang_between(c, b, a))

    def test_sketch_quadrange(self):
        a, b, c, d = nm.sketch_quadrangle([])
        check.is_true(Line(a, c).diff_side(b, d))
        check.is_true(Line(b, d).diff_side(a, c))

    def test_sketch_r_trapezoid(self):
        a, b, c, d = nm.sketch_r_trapezoid([])
        check.is_true(Line(a, b).is_perp(Line(a, d)))
        check.is_true(Line(a, b).is_parallel(Line(c, d)))
        check.is_true(Line(a, c).diff_side(b, d))
        check.is_true(Line(b, d).diff_side(a, c))

    def test_sketch_r_triangle(self):
        a, b, c = nm.sketch_r_triangle([])
        check.is_true(Line(a, b).is_perp(Line(a, c)))

    def test_sketch_rectangle(self):
        a, b, c, d = nm.sketch_rectangle([])
        check.is_true(Line(a, b).is_perp(Line(b, c)))
        check.is_true(Line(b, c).is_perp(Line(c, d)))
        check.is_true(Line(c, d).is_perp(Line(d, a)))

    def test_sketch_reflect(self):
        a, b, c = random_points(3)
        x = nm.sketch_reflect([a, b, c])
        check.is_true(Line(a, x).is_perp(Line(b, c)))
        check.almost_equal(x.distance(Line(b, c)), a.distance(Line(b, c)))

    def test_sketch_risos(self):
        a, b, c = nm.sketch_risos([])
        check.almost_equal(a.distance(b), a.distance(c))
        check.is_true(Line(a, b).is_perp(Line(a, c)))

    def test_sketch_rotaten90(self):
        a, b = random_points(2)
        x = nm.sketch_rotaten90([a, b])
        check.almost_equal(a.distance(x), a.distance(b))
        check.is_true(Line(a, x).is_perp(Line(a, b)))
        d = Point(0.0, 0.0)
        e = Point(0.0, 1.0)
        f = Point(1.0, 0.0)
        check.almost_equal(ang_between(d, e, f), ang_between(a, b, x))

    def test_sketch_rotatep90(self):
        a, b = random_points(2)
        x = nm.sketch_rotatep90([a, b])
        check.almost_equal(a.distance(x), a.distance(b))
        check.is_true(Line(a, x).is_perp(Line(a, b)))
        d = Point(0.0, 0.0)
        e = Point(0.0, 1.0)
        f = Point(1.0, 0.0)
        check.almost_equal(ang_between(d, f, e), ang_between(a, b, x))

    def test_sketch_s_angle(self):
        a, b = random_points(2)
        y = unif(0.0, np.pi)
        bx = nm.sketch_s_angle([a, b, y / np.pi * 180])
        check.is_instance(bx, HalfLine)
        check.equal(bx.tail, b)
        x = bx.head

        d = Point(1.0, 0.0)
        e = Point(0.0, 0.0)
        f = Point(np.cos(y), np.sin(y))
        check.almost_equal(ang_between(e, d, f), ang_between(b, a, x))

    def test_sketch_shift(self):
        a, b, c = random_points(3)
        x = nm.sketch_shift([a, b, c])
        check.is_true((b - a).close(x - c))

    def test_sketch_square(self):
        a, b = random_points(2)
        c, d = nm.sketch_square([a, b])
        check.is_true(Line(a, b).is_perp(Line(b, c)))
        check.is_true(Line(b, c).is_perp(Line(c, d)))
        check.is_true(Line(c, d).is_perp(Line(d, a)))
        check.almost_equal(a.distance(b), b.distance(c))

    def test_sketch_isquare(self):
        a, b, c, d = nm.sketch_isquare([])
        check.is_true(Line(a, b).is_perp(Line(b, c)))
        check.is_true(Line(b, c).is_perp(Line(c, d)))
        check.is_true(Line(c, d).is_perp(Line(d, a)))
        check.almost_equal(a.distance(b), b.distance(c))

    def test_sketch_trapezoid(self):
        a, b, c, d = nm.sketch_trapezoid([])
        check.is_true(Line(a, b).is_parallel(Line(c, d)))
        check.is_true(Line(a, c).diff_side(b, d))
        check.is_true(Line(b, d).diff_side(a, c))

    def test_sketch_triangle(self):
        a, b, c = nm.sketch_triangle([])
        check.is_false(check_coll([a, b, c]))

    def test_sketch_triangle12(self):
        a, b, c = nm.sketch_triangle12([])
        check.almost_equal(a.distance(b) * 2, a.distance(c))

    def test_sketch_trisect(self):
        a, b, c = random_points(3)
        x, y = nm.sketch_trisect([a, b, c])
        check.almost_equal(ang_between(b, a, x), ang_between(b, x, y))
        check.almost_equal(ang_between(b, x, y), ang_between(b, y, c))
        check.almost_equal(ang_between(b, a, x) * 3, ang_between(b, a, c))

    def test_sketch_trisegment(self):
        a, b = random_points(2)
        x, y = nm.sketch_trisegment([a, b])
        check.almost_equal(a.distance(x) + x.distance(y) + y.distance(b), a.distance(b))
        check.almost_equal(a.distance(x), x.distance(y))
        check.almost_equal(x.distance(y), y.distance(b))



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Numerical representation of geometry."""

from __future__ import annotations
from typing import Any, Optional, Union

import math

import matplotlib
from matplotlib import pyplot as plt
import matplotlib.colors as mcolors

import numpy as np
from numpy.random import uniform as unif


import newclid.geometry as gm

ATOM = 1e-12


# Some variables are there for better code reading.


# Naming in geometry is a little different
# we stick to geometry naming to better read the code.


class Point:
    """Numerical point."""

    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __lt__(self, other: Point) -> bool:
        return (self.x, self.y) < (other.x, other.y)

    def __gt__(self, other: Point) -> bool:
        return (self.x, self.y) > (other.x, other.y)

    def __add__(self, p: Point) -> Point:
        return Point(self.x + p.x, self.y + p.y)

    def __sub__(self, p: Point) -> Point:
        return Point(self.x - p.x, self.y - p.y)

    def __mul__(self, f: float) -> Point:
        return Point(self.x * f, self.y * f)

    def __rmul__(self, f: float) -> Point:
        return self * f

    def __truediv__(self, f: float) -> Point:
        return Point(self.x / f, self.y / f)

    def __floordiv__(self, f: float) -> Point:
        div = self / f  # true div
        return Point(int(div.x), int(div.y))

    def __str__(self) -> str:
        return "P({},{})".format(self.x, self.y)

    def close(self, point: Point, tol: float = 1e-12) -> bool:
        return abs(self.x - point.x) < tol and abs(self.y - point.y) < tol

    def midpoint(self, p: Point) -> Point:
        return Point(0.5 * (self.x + p.x), 0.5 * (self.y + p.y))

    def distance(self, p: Union[Point, Line, Circle]) -> float:
        if isinstance(p, Line):
            return p.distance(self)
        if isinstance(p, Circle):
            return abs(p.radius - self.distance(p.center))
        dx = self.x - p.x
        dy = self.y - p.y
        return np.sqrt(dx * dx + dy * dy)

    def distance2(self, p: Point) -> float:
        if isinstance(p, Line):
            return p.distance(self)
        dx = self.x - p.x
        dy = self.y - p.y
        return dx * dx + dy * dy

    def rotatea(self, ang: float) -> Point:
        sinb, cosb = np.sin(ang), np.cos(ang)
        return self.rotate(sinb, cosb)

    def rotate(self, sinb: float, cosb: float) -> Point:
        x, y = self.x, self.y
        return Point(x * cosb - y * sinb, x * sinb + y * cosb)

    def flip(self) -> Point:
        return Point(-self.x, self.y)

    def perpendicular_line(self, line: Line) -> Line:
        return line.perpendicular_line(self)

    def foot(self, line: Line) -> Point:
        if isinstance(line, Line):
            perpendicular_line = line.perpendicular_line(self)
            return line_line_intersection(perpendicular_line, line)
        elif isinstance(line, Circle):
            c, r = line.center, line.radius
            return c + (self - c) * r / self.distance(c)
        raise ValueError("Dropping foot to weird type {}".format(type(line)))

    def parallel_line(self, line: Line) -> Line:
        return line.parallel_line(self)

    def norm(self) -> float:
        return np.sqrt(self.x**2 + self.y**2)

    def cos(self, other: Point) -> float:
        x, y = self.x, self.y
        a, b = other.x, other.y
        return (x * a + y * b) / self.norm() / other.norm()

    def dot(self, other: Point) -> float:
        return self.x * other.x + self.y * other.y

    def sign(self, line: Line) -> int:
        return line.sign(self)

    def is_same(self, other: Point) -> bool:
        return self.distance(other) <= ATOM


class Line:
    """Numerical line."""

    def __init__(
        self,
        p1: Point = None,
        p2: Point = None,
        coefficients: tuple[int, int, int] = None,
    ):
        if p1 is None and p2 is None and coefficients is None:
            self.coefficients = None, None, None
            return

        a, b, c = coefficients or (
            p1.y - p2.y,
            p2.x - p1.x,
            p1.x * p2.y - p2.x * p1.y,
        )

        # Make sure a is always positive (or always negative for that matter)
        # With a == 0, Assuming a = +epsilon > 0
        # Then b such that ax + by = 0 with y>0 should be negative.
        if a < 0.0 or a == 0.0 and b > 0.0:
            a, b, c = -a, -b, -c

        self.coefficients = a, b, c

    def parallel_line(self, p: Point) -> Line:
        a, b, _ = self.coefficients
        return Line(coefficients=(a, b, -a * p.x - b * p.y))

    def perpendicular_line(self, p: Point) -> Line:
        a, b, _ = self.coefficients
        return Line(p, p + Point(a, b))

    def greater_than(self, other: Line) -> bool:
        a, b, _ = self.coefficients
        x, y, _ = other.coefficients
        # b/a > y/x
        return b * x > a * y

    def __gt__(self, other: Line) -> bool:
        return self.greater_than(other)

    def __lt__(self, other: Line) -> bool:
        return other.greater_than(self)

    def same(self, other: Line) -> bool:
        a, b, c = self.coefficients
        x, y, z = other.coefficients
        return close_enough(a * y, b * x) and close_enough(b * z, c * y)

    def equal(self, other: Line) -> bool:
        a, b, _ = self.coefficients
        x, y, _ = other.coefficients
        # b/a == y/x
        return b * x == a * y

    def less_than(self, other: Line) -> bool:
        a, b, _ = self.coefficients
        x, y, _ = other.coefficients
        # b/a > y/x
        return b * x < a * y

    def intersect(self, obj: Union[Line, Circle]) -> tuple[Point, ...]:
        if isinstance(obj, Line):
            return line_line_intersection(self, obj)
        if isinstance(obj, Circle):
            return line_circle_intersection(self, obj)

    def distance(self, p: Point) -> float:
        a, b, c = self.coefficients
        return abs(self(p.x, p.y)) / math.sqrt(a * a + b * b)

    def __call__(self, x: Point, y: Point = None) -> float:
        if isinstance(x, Point) and y is None:
            return self(x.x, x.y)
        a, b, c = self.coefficients
        return x * a + y * b + c

    def is_parallel(self, other: Line) -> bool:
        a, b, _ = self.coefficients
        x, y, _ = other.coefficients
        return abs(a * y - b * x) < ATOM

    def is_perp(self, other: Line) -> bool:
        a, b, _ = self.coefficients
        x, y, _ = other.coefficients
        return abs(a * x + b * y) < ATOM

    def cross(self, other: Line) -> float:
        a, b, _ = self.coefficients
        x, y, _ = other.coefficients
        return a * y - b * x

    def dot(self, other: Line) -> float:
        a, b, _ = self.coefficients
        x, y, _ = other.coefficients
        return a * x + b * y

    def point_at(self, x: float = None, y: float = None) -> Optional[Point]:
        """Get a point on line closest to (x, y)."""
        a, b, c = self.coefficients
        # ax + by + c = 0
        if x is None and y is not None:
            if a != 0:
                return Point((-c - b * y) / a, y)
            else:
                return None
        elif x is not None and y is None:
            if b != 0:
                return Point(x, (-c - a * x) / b)
            else:
                return None
        elif x is not None and y is not None:
            if a * x + b * y + c == 0.0:
                return Point(x, y)
        return None

    def diff_side(self, p1: Point, p2: Point) -> Optional[bool]:
        d1 = self(p1.x, p1.y)
        d2 = self(p2.x, p2.y)
        if d1 == 0 or d2 == 0:
            return None
        return d1 * d2 < 0

    def same_side(self, p1: Point, p2: Point) -> Optional[bool]:
        d1 = self(p1.x, p1.y)
        d2 = self(p2.x, p2.y)
        if d1 == 0 or d2 == 0:
            return None
        return d1 * d2 > 0

    def sign(self, point: Point) -> int:
        s = self(point.x, point.y)
        if s > 0:
            return 1
        elif s < 0:
            return -1
        return 0

    def is_same(self, other: Line) -> bool:
        a, b, c = self.coefficients
        x, y, z = other.coefficients
        return abs(a * y - b * x) <= ATOM and abs(b * z - c * y) <= ATOM

    def sample_within(self, points: list[Point], n: int = 5) -> list[Point]:
        """Sample a point within the boundary of points."""
        center = sum(points, Point(0.0, 0.0)) * (1.0 / len(points))
        radius = max([p.distance(center) for p in points])
        if close_enough(center.distance(self), radius):
            center = center.foot(self)
        a, b = line_circle_intersection(self, Circle(center.foot(self), radius))

        result = None
        best = -1.0
        for _ in range(n):
            rand = unif(0.0, 1.0)
            x = a + (b - a) * rand
            mind = min([x.distance(p) for p in points])
            if mind > best:
                best = mind
                result = x

        return [result]


class InvalidLineIntersectError(Exception):
    pass


class HalfLine(Line):
    """Numerical ray."""

    def __init__(self, tail: Point, head: Point):
        self.line = Line(tail, head)
        self.coefficients = self.line.coefficients
        self.tail = tail
        self.head = head

    def intersect(self, obj: Union[Line, HalfLine, Circle, HoleCircle]) -> Point:
        if isinstance(obj, (HalfLine, Line)):
            return line_line_intersection(self.line, obj)

        exclude = [self.tail]
        if isinstance(obj, HoleCircle):
            exclude += [obj.hole]

        a, b = line_circle_intersection(self.line, obj)
        if any([a.close(x) for x in exclude]):
            return b
        if any([b.close(x) for x in exclude]):
            return a

        v = self.head - self.tail
        va = a - self.tail
        vb = b - self.tail
        if v.dot(va) > 0:
            return a
        if v.dot(vb) > 0:
            return b
        raise InvalidLineIntersectError()

    def sample_within(self, points: list[Point], n: int = 5) -> list[Point]:
        center = sum(points, Point(0.0, 0.0)) * (1.0 / len(points))
        radius = max([p.distance(center) for p in points])
        if close_enough(center.distance(self.line), radius):
            center = center.foot(self)
        a, b = line_circle_intersection(self, Circle(center.foot(self), radius))

        if (a - self.tail).dot(self.head - self.tail) > 0:
            a, b = self.tail, a
        else:
            a, b = self.tail, b

        result = None
        best = -1.0
        for _ in range(n):
            x = a + (b - a) * unif(0.0, 1.0)
            mind = min([x.distance(p) for p in points])
            if mind > best:
                best = mind
                result = x

        return [result]


def _perpendicular_bisector(p1: Point, p2: Point) -> Line:
    midpoint = (p1 + p2) * 0.5
    return Line(midpoint, midpoint + Point(p2.y - p1.y, p1.x - p2.x))


def same_sign(a: Point, b: Point, c: Point, d: Point, e: Point, f: Point) -> bool:
    a, b, c, d, e, f = map(lambda p: p.sym, [a, b, c, d, e, f])
    ab, cb = a - b, c - b
    de, fe = d - e, f - e
    return (ab.x * cb.y - ab.y * cb.x) * (de.x * fe.y - de.y * fe.x) > 0


class Circle:
    """Numerical circle."""

    def __init__(
        self,
        center: Optional[Point] = None,
        radius: Optional[float] = None,
        p1: Optional[Point] = None,
        p2: Optional[Point] = None,
        p3: Optional[Point] = None,
    ):
        if not center:
            if not (p1 and p2 and p3):
                self.center = self.radius = self.r2 = None
                return
                # raise ValueError('Circle without center need p1 p2 p3')

            l12 = _perpendicular_bisector(p1, p2)
            l23 = _perpendicular_bisector(p2, p3)
            center = line_line_intersection(l12, l23)

        self.center = center
        self.a, self.b = center.x, center.y

        if not radius:
            if not (p1 or p2 or p3):
                raise ValueError("Circle needs radius or p1 or p2 or p3")
            p = p1 or p2 or p3
            self.r2 = (self.a - p.x) ** 2 + (self.b - p.y) ** 2
            self.radius = math.sqrt(self.r2)
        else:
            self.radius = radius
            self.r2 = radius * radius

    def intersect(self, obj: Union[Line, Circle]) -> tuple[Point, ...]:
        if isinstance(obj, Line):
            return obj.intersect(self)
        if isinstance(obj, Circle):
            return circle_circle_intersection(self, obj)

    def sample_within(self, points: list[Point], n: int = 5) -> list[Point]:
        """Sample a point within the boundary of points."""
        result = None
        best = -1.0
        for _ in range(n):
            ang = unif(0.0, 2.0) * np.pi
            x = self.center + Point(np.cos(ang), np.sin(ang)) * self.radius
            mind = min([x.distance(p) for p in points])
            if mind > best:
                best = mind
                result = x

        return [result]


class HoleCircle(Circle):
    """Numerical circle with a missing point."""

    def __init__(self, center: Point, radius: float, hole: Point):
        super().__init__(center, radius)
        self.hole = hole

    def intersect(self, obj: Union[Line, HalfLine, Circle, HoleCircle]) -> Point:
        if isinstance(obj, Line):
            a, b = line_circle_intersection(obj, self)
            if a.close(self.hole):
                return b
            return a
        if isinstance(obj, HalfLine):
            return obj.intersect(self)
        if isinstance(obj, Circle):
            a, b = circle_circle_intersection(obj, self)
            if a.close(self.hole):
                return b
            return a
        if isinstance(obj, HoleCircle):
            a, b = circle_circle_intersection(obj, self)
            if a.close(self.hole) or a.close(obj.hole):
                return b
            return a


def solve_quad(a: float, b: float, c: float) -> tuple[float, float]:
    """Solve a x^2 + bx + c = 0."""
    a = 2 * a
    d = b * b - 2 * a * c
    if d < 0:
        return None  # the caller should expect this result.

    y = math.sqrt(d)
    return (-b - y) / a, (-b + y) / a


def circle_circle_intersection(c1: Circle, c2: Circle) -> tuple[Point, Point]:
    """Returns a pair of Points as intersections of c1 and c2."""
    # circle 1: (x0, y0), radius r0
    # circle 2: (x1, y1), radius r1
    x0, y0, r0 = c1.a, c1.b, c1.radius
    x1, y1, r1 = c2.a, c2.b, c2.radius

    d = math.sqrt((x1 - x0) ** 2 + (y1 - y0) ** 2)
    if d == 0:
        raise InvalidQuadSolveError()

    a = (r0**2 - r1**2 + d**2) / (2 * d)
    h = r0**2 - a**2
    if h < 0:
        raise InvalidQuadSolveError()
    h = np.sqrt(h)
    x2 = x0 + a * (x1 - x0) / d
    y2 = y0 + a * (y1 - y0) / d
    x3 = x2 + h * (y1 - y0) / d
    y3 = y2 - h * (x1 - x0) / d
    x4 = x2 - h * (y1 - y0) / d
    y4 = y2 + h * (x1 - x0) / d

    return Point(x3, y3), Point(x4, y4)


class InvalidQuadSolveError(Exception):
    pass


def line_circle_intersection(line: Line, circle: Circle) -> tuple[Point, Point]:
    """Returns a pair of points as intersections of line and circle."""
    a, b, c = line.coefficients
    r = float(circle.radius)
    center = circle.center
    p, q = center.x, center.y

    if b == 0:
        x = -c / a
        x_p = x - p
        x_p2 = x_p * x_p
        y = solve_quad(1, -2 * q, q * q + x_p2 - r * r)
        if y is None:
            raise InvalidQuadSolveError()
        y1, y2 = y
        return (Point(x, y1), Point(x, y2))

    if a == 0:
        y = -c / b
        y_q = y - q
        y_q2 = y_q * y_q
        x = solve_quad(1, -2 * p, p * p + y_q2 - r * r)
        if x is None:
            raise InvalidQuadSolveError()
        x1, x2 = x
        return (Point(x1, y), Point(x2, y))

    c_ap = c + a * p
    a2 = a * a
    y = solve_quad(
        a2 + b * b, 2 * (b * c_ap - a2 * q), c_ap * c_ap + a2 * (q * q - r * r)
    )
    if y is None:
        raise InvalidQuadSolveError()
    y1, y2 = y

    return Point(-(b * y1 + c) / a, y1), Point(-(b * y2 + c) / a, y2)


def _check_between(a: Point, b: Point, c: Point) -> bool:
    """Whether a is between b & c."""
    return (a - b).dot(c - b) > 0 and (a - c).dot(b - c) > 0


def circle_segment_intersect(circle: Circle, p1: Point, p2: Point) -> list[Point]:
    line = Line(p1, p2)
    px, py = line_circle_intersection(line, circle)

    result = []
    if _check_between(px, p1, p2):
        result.append(px)
    if _check_between(py, p1, p2):
        result.append(py)
    return result


def line_segment_intersection(line: Line, A: Point, B: Point) -> Point:
    a, b, c = line.coefficients
    x1, y1, x2, y2 = A.x, A.y, B.x, B.y
    dx, dy = x2 - x1, y2 - y1
    alpha = (-c - a * x1 - b * y1) / (a * dx + b * dy)
    return Point(x1 + alpha * dx, y1 + alpha * dy)


def line_line_intersection(line_1: Line, line_2: Line) -> Point:
    a1, b1, c1 = line_1.coefficients
    a2, b2, c2 = line_2.coefficients
    # a1x + b1y + c1 = 0
    # a2x + b2y + c2 = 0
    d = a1 * b2 - a2 * b1
    if d == 0:
        raise InvalidLineIntersectError
    return Point((c2 * b1 - c1 * b2) / d, (c1 * a2 - c2 * a1) / d)


def check_too_close(
    newpoints: list[Point], points: list[Point], tol: int = 0.1
) -> bool:
    if not points:
        return False
    avg = sum(points, Point(0.0, 0.0)) * 1.0 / len(points)
    mindist = min([p.distance(avg) for p in points])
    for p0 in newpoints:
        for p1 in points:
            if p0.distance(p1) < tol * mindist:
                return True
    return False


def check_too_far(newpoints: list[Point], points: list[Point], tol: int = 4) -> bool:
    if len(points) < 2:
        return False
    avg = sum(points, Point(0.0, 0.0)) * 1.0 / len(points)
    maxdist = max([p.distance(avg) for p in points])
    for p in newpoints:
        if p.distance(avg) > maxdist * tol:
            return True
    return False


def check_aconst(args: list[Point]) -> bool:
    a, b, c, d, num, den = args
    d = d + a - c
    ang = ang_between(a, b, d)
    if ang < 0:
        ang += np.pi
    return close_enough(ang, num * np.pi / den)


def check(name: str, args: list[Union[gm.Point, Point]]) -> bool:
    """Numerical check."""
    if name == "eqangle6":
        name = "eqangle"
    elif name == "eqratio6":
        name = "eqratio"
    elif name in ["simtri2", "simtri*"]:
        name = "simtri"
    elif name in ["contri2", "contri*"]:
        name = "contri"
    elif name == "para":
        name = "para_or_coll"
    elif name == "on_line":
        name = "coll"
    elif name in ["rcompute", "acompute"]:
        return True
    elif name in ["fixl", "fixc", "fixb", "fixt", "fixp"]:
        return True

    fn_name = "check_" + name
    if fn_name not in globals():
        return None

    fun = globals()["check_" + name]
    args = [p.num if isinstance(p, gm.Point) else p for p in args]
    return fun(args)


def check_circle(points: list[Point]) -> bool:
    if len(points) != 4:
        return False
    o, a, b, c = points
    oa, ob, oc = o.distance(a), o.distance(b), o.distance(c)
    return close_enough(oa, ob) and close_enough(ob, oc)


def check_coll(points: list[Point]) -> bool:
    a, b = points[:2]
    line = Line(a, b)
    for p in points[2:]:
        if abs(line(p.x, p.y)) > ATOM:
            return False
    return True


def check_ncoll(points: list[Point]) -> bool:
    return not check_coll(points)


def check_sameside(points: list[Point]) -> bool:
    b, a, c, y, x, z = points
    # whether b is to the same side of a & c as y is to x & z
    ba = b - a
    bc = b - c
    yx = y - x
    yz = y - z
    return ba.dot(bc) * yx.dot(yz) > 0


def check_para_or_coll(points: list[Point]) -> bool:
    return check_para(points) or check_coll(points)


def check_para(points: list[Point]) -> bool:
    a, b, c, d = points
    ab = Line(a, b)
    cd = Line(c, d)
    if ab.same(cd):
        return False
    return ab.is_parallel(cd)


def check_perp(points: list[Point]) -> bool:
    a, b, c, d = points
    ab = Line(a, b)
    cd = Line(c, d)
    return ab.is_perp(cd)


def check_cyclic(points: list[Point]) -> bool:
    points = list(set(points))
    a, b, c, *ps = points
    circle = Circle(p1=a, p2=b, p3=c)
    for d in ps:
        if not close_enough(d.distance(circle.center), circle.radius):
            return False
    return True


def bring_together(
    a: Point, b: Point, c: Point, d: Point
) -> tuple[Point, Point, Point, Point]:
    ab = Line(a, b)
    cd = Line(c, d)
    x = line_line_intersection(ab, cd)
    unit = Circle(center=x, radius=1.0)
    y, _ = line_circle_intersection(ab, unit)
    z, _ = line_circle_intersection(cd, unit)
    return x, y, x, z


def same_clock(a: Point, b: Point, c: Point, d: Point, e: Point, f: Point) -> bool:
    ba = b - a
    cb = c - b
    ed = e - d
    fe = f - e
    return (ba.x * cb.y - ba.y * cb.x) * (ed.x * fe.y - ed.y * fe.x) > 0


def check_const_angle(points: list[Point]) -> bool:
    """Check if the angle is equal to the given constant."""
    a, b, c, d, m, n = points
    a, b, c, d = bring_together(a, b, c, d)
    ba = b - a
    dc = d - c

    a3 = np.arctan2(ba.y, ba.x)
    a4 = np.arctan2(dc.y, dc.x)
    y = a3 - a4

    return close_enough(m / n % 1, y / np.pi % 1)


def check_eqangle(points: list[Point]) -> bool:
    """Check if 8 points make 2 equal angles."""
    a, b, c, d, e, f, g, h = points

    ab = Line(a, b)
    cd = Line(c, d)
    ef = Line(e, f)
    gh = Line(g, h)

    if ab.is_parallel(cd):
        return ef.is_parallel(gh)
    if ef.is_parallel(gh):
        return ab.is_parallel(cd)

    a, b, c, d = bring_together(a, b, c, d)
    e, f, g, h = bring_together(e, f, g, h)

    ba = b - a
    dc = d - c
    fe = f - e
    hg = h - g

    sameclock = (ba.x * dc.y - ba.y * dc.x) * (fe.x * hg.y - fe.y * hg.x) > 0
    if not sameclock:
        ba = ba * -1.0

    a1 = np.arctan2(fe.y, fe.x)
    a2 = np.arctan2(hg.y, hg.x)
    x = a1 - a2

    a3 = np.arctan2(ba.y, ba.x)
    a4 = np.arctan2(dc.y, dc.x)
    y = a3 - a4

    xy = (x - y) % (2 * np.pi)
    return close_enough(xy, 0, tol=1e-11) or close_enough(xy, 2 * np.pi, tol=1e-11)


def check_eqratio(points: list[Point]) -> bool:
    a, b, c, d, e, f, g, h = points
    ab = a.distance(b)
    cd = c.distance(d)
    ef = e.distance(f)
    gh = g.distance(h)
    return close_enough(ab * gh, cd * ef)


def check_cong(points: list[Point]) -> bool:
    a, b, c, d = points
    return close_enough(a.distance(b), c.distance(d))


def check_midp(points: list[Point]) -> bool:
    a, b, c = points
    return check_coll(points) and close_enough(a.distance(b), a.distance(c))


def check_simtri(points: list[Point]) -> bool:
    """Check if 6 points make a pair of similar triangles."""
    a, b, c, x, y, z = points
    ab = a.distance(b)
    bc = b.distance(c)
    ca = c.distance(a)
    xy = x.distance(y)
    yz = y.distance(z)
    zx = z.distance(x)
    tol = 1e-9
    return close_enough(ab * yz, bc * xy, tol) and close_enough(bc * zx, ca * yz, tol)


def check_contri(points: list[Point]) -> bool:
    a, b, c, x, y, z = points
    ab = a.distance(b)
    bc = b.distance(c)
    ca = c.distance(a)
    xy = x.distance(y)
    yz = y.distance(z)
    zx = z.distance(x)
    tol = 1e-9
    return (
        close_enough(ab, xy, tol)
        and close_enough(bc, yz, tol)
        and close_enough(ca, zx, tol)
    )


def check_ratio(points: list[Point]) -> bool:
    a, b, c, d, m, n = points
    ab = a.distance(b)
    cd = c.distance(d)
    return close_enough(ab * n, cd * m)


def draw_angle(
    ax: matplotlib.axes.Axes,
    head: Point,
    p1: Point,
    p2: Point,
    color: Any = "red",
    alpha: float = 0.5,
    frac: float = 1.0,
) -> None:
    """Draw an angle on plt ax."""
    d1 = p1 - head
    d2 = p2 - head

    a1 = np.arctan2(float(d1.y), float(d1.x))
    a2 = np.arctan2(float(d2.y), float(d2.x))
    a1, a2 = a1 * 180 / np.pi, a2 * 180 / np.pi
    a1, a2 = a1 % 360, a2 % 360

    if a1 > a2:
        a1, a2 = a2, a1

    if a2 - a1 > 180:
        a1, a2 = a2, a1

    b1, b2 = a1, a2
    if b1 > b2:
        b2 += 360
    d = b2 - b1
    # if d >= 90:
    #   return

    scale = min(2.0, 90 / d)
    scale = max(scale, 0.4)
    fov = matplotlib.patches.Wedge(
        (float(head.x), float(head.y)),
        unif(0.075, 0.125) * scale * frac,
        a1,
        a2,
        color=color,
        alpha=alpha,
    )
    ax.add_artist(fov)


def naming_position(
    ax: matplotlib.axes.Axes, p: Point, lines: list[Line], circles: list[Circle]
) -> tuple[float, float]:
    """Figure out a good naming position on the drawing."""
    _ = ax
    r = 0.08
    c = Circle(center=p, radius=r)
    avoid = []
    for p1, p2 in lines:
        try:
            avoid.extend(circle_segment_intersect(c, p1, p2))
        except InvalidQuadSolveError:
            continue
    for x in circles:
        try:
            avoid.extend(circle_circle_intersection(c, x))
        except InvalidQuadSolveError:
            continue

    if not avoid:
        return [p.x + 0.01, p.y + 0.01]

    angs = sorted([ang_of(p, a) for a in avoid])
    angs += [angs[0] + 2 * np.pi]
    angs = [(angs[i + 1] - a, a) for i, a in enumerate(angs[:-1])]

    d, a = max(angs)
    ang = a + d / 2

    name_pos = p + Point(np.cos(ang), np.sin(ang)) * r

    x, y = (name_pos.x - r / 1.5, name_pos.y - r / 1.5)
    return x, y


def draw_point(
    ax: matplotlib.axes.Axes,
    p: Point,
    name: str,
    lines: list[Line],
    circles: list[Circle],
    color: Any = "white",
    size: float = 15,
) -> None:
    """draw a point."""
    ax.scatter(p.x, p.y, color=color, s=size)

    if color == "white":
        color = "lightgreen"
    else:
        color = "grey"

    name = name.upper()
    if len(name) > 1:
        name = name[0] + "_" + name[1:]

    ax.annotate(name, naming_position(ax, p, lines, circles), color=color, fontsize=15)


def _draw_line(
    ax: matplotlib.axes.Axes,
    p1: Point,
    p2: Point,
    color: Any = "white",
    lw: float = 1.2,
    alpha: float = 0.8,
) -> None:
    """Draw a line in matplotlib."""
    ls = "-"
    if color == "--":
        color = "black"
        ls = "--"

    lx, ly = (p1.x, p2.x), (p1.y, p2.y)
    ax.plot(lx, ly, color=color, lw=lw, alpha=alpha, ls=ls)


def draw_line(
    ax: matplotlib.axes.Axes, line: Line, color: Any = "white"
) -> tuple[Point, Point]:
    """Draw a line."""
    points = line.neighbors(gm.Point)
    if len(points) <= 1:
        return

    points = [p.num for p in points]
    p1, p2 = points[:2]

    pmin, pmax = (p1, 0.0), (p2, (p2 - p1).dot(p2 - p1))

    for p in points[2:]:
        v = (p - p1).dot(p2 - p1)
        if v < pmin[1]:
            pmin = p, v
        if v > pmax[1]:
            pmax = p, v

    p1, p2 = pmin[0], pmax[0]
    _draw_line(ax, p1, p2, color=color)
    return p1, p2


def _draw_circle(
    ax: matplotlib.axes.Axes, c: Circle, color: Any = "cyan", lw: float = 1.2
) -> None:
    ls = "-"
    if color == "--":
        color = "black"
        ls = "--"

    ax.add_patch(
        plt.Circle(
            (c.center.x, c.center.y),
            c.radius,
            color=color,
            alpha=0.8,
            fill=False,
            lw=lw,
            ls=ls,
        )
    )


def draw_circle(
    ax: matplotlib.axes.Axes, circle: Circle, color: Any = "cyan"
) -> Circle:
    """Draw a circle."""
    if circle.num is not None:
        circle = circle.num
    else:
        points = circle.neighbors(gm.Point)
        if len(points) <= 2:
            return
        points = [p.num for p in points]
        p1, p2, p3 = points[:3]
        circle = Circle(p1=p1, p2=p2, p3=p3)

    _draw_circle(ax, circle, color)
    return circle


def mark_segment(
    ax: matplotlib.axes.Axes, p1: Point, p2: Point, color: Any, alpha: float
) -> None:
    _ = alpha
    x, y = (p1.x + p2.x) / 2, (p1.y + p2.y) / 2
    ax.scatter(x, y, color=color, alpha=1.0, marker="o", s=50)


def highlight_angle(
    ax: matplotlib.axes.Axes,
    a: Point,
    b: Point,
    c: Point,
    d: Point,
    color: Any,
    alpha: float,
) -> None:
    """Highlight an angle between ab and cd with (color, alpha)."""
    try:
        a, b, c, d = bring_together(a, b, c, d)
    except (InvalidLineIntersectError, InvalidQuadSolveError):
        return
    draw_angle(ax, a, b, d, color=color, alpha=alpha, frac=1.0)


def highlight(
    ax: matplotlib.axes.Axes,
    name: str,
    args: list[gm.Point],
    lcolor: Any,
    color1: Any,
    color2: Any,
) -> None:
    """Draw highlights."""
    args = list(map(lambda x: x.num if isinstance(x, gm.Point) else x, args))

    if name == "cyclic":
        a, b, c, d = args
        _draw_circle(ax, Circle(p1=a, p2=b, p3=c), color=color1, lw=2.0)
    if name == "coll":
        a, b, c = args
        a, b = max(a, b, c), min(a, b, c)
        _draw_line(ax, a, b, color=color1, lw=2.0)
    if name == "para":
        a, b, c, d = args
        _draw_line(ax, a, b, color=color1, lw=2.0)
        _draw_line(ax, c, d, color=color2, lw=2.0)
    if name == "eqangle":
        a, b, c, d, e, f, g, h = args

        x = line_line_intersection(Line(a, b), Line(c, d))
        if b.distance(x) > a.distance(x):
            a, b = b, a
        if d.distance(x) > c.distance(x):
            c, d = d, c
        a, b, d = x, a, c

        y = line_line_intersection(Line(e, f), Line(g, h))
        if f.distance(y) > e.distance(y):
            e, f = f, e
        if h.distance(y) > g.distance(y):
            g, h = h, g
        e, f, h = y, e, g

        _draw_line(ax, a, b, color=lcolor, lw=2.0)
        _draw_line(ax, a, d, color=lcolor, lw=2.0)
        _draw_line(ax, e, f, color=lcolor, lw=2.0)
        _draw_line(ax, e, h, color=lcolor, lw=2.0)
        if color1 == "--":
            color1 = "red"
        draw_angle(ax, a, b, d, color=color1, alpha=0.5)
        if color2 == "--":
            color2 = "red"
        draw_angle(ax, e, f, h, color=color2, alpha=0.5)
    if name == "perp":
        a, b, c, d = args
        _draw_line(ax, a, b, color=color1, lw=2.0)
        _draw_line(ax, c, d, color=color1, lw=2.0)
    if name == "ratio":
        a, b, c, d, m, n = args
        _draw_line(ax, a, b, color=color1, lw=2.0)
        _draw_line(ax, c, d, color=color2, lw=2.0)
    if name == "cong":
        a, b, c, d = args
        _draw_line(ax, a, b, color=color1, lw=2.0)
        _draw_line(ax, c, d, color=color2, lw=2.0)
    if name == "midp":
        m, a, b = args
        _draw_line(ax, a, m, color=color1, lw=2.0, alpha=0.5)
        _draw_line(ax, b, m, color=color2, lw=2.0, alpha=0.5)
    if name == "eqratio":
        a, b, c, d, m, n, p, q = args
        _draw_line(ax, a, b, color=color1, lw=2.0, alpha=0.5)
        _draw_line(ax, c, d, color=color2, lw=2.0, alpha=0.5)
        _draw_line(ax, m, n, color=color1, lw=2.0, alpha=0.5)
        _draw_line(ax, p, q, color=color2, lw=2.0, alpha=0.5)


HCOLORS = None


def _draw(
    ax: matplotlib.axes.Axes,
    points: list[gm.Point],
    lines: list[gm.Line],
    circles: list[gm.Circle],
    goal: Any,
    equals: list[tuple[Any, Any]],
    highlights: list[tuple[str, list[gm.Point]]],
):
    """Draw everything."""
    colors = ["red", "green", "blue", "orange", "magenta", "purple"]
    pcolor = "black"
    lcolor = "black"
    ccolor = "grey"
    if get_theme() == "dark":
        pcolor, lcolor, ccolor = "white", "white", "cyan"
    elif get_theme() == "light":
        pcolor, lcolor, ccolor = "black", "black", "blue"
    elif get_theme() == "grey":
        pcolor, lcolor, ccolor = "black", "black", "grey"
        colors = ["grey"]

    line_boundaries = []
    for line in lines:
        p1, p2 = draw_line(ax, line, color=lcolor)
        line_boundaries.append((p1, p2))
    circles = [draw_circle(ax, c, color=ccolor) for c in circles]

    for p in points:
        draw_point(ax, p.num, p.name, line_boundaries, circles, color=pcolor)

    if equals:
        for i, segs in enumerate(equals["segments"]):
            color = colors[i % len(colors)]
            for a, b in segs:
                mark_segment(ax, a, b, color, 0.5)

        for i, angs in enumerate(equals["angles"]):
            color = colors[i % len(colors)]
            for a, b, c, d in angs:
                highlight_angle(ax, a, b, c, d, color, 0.5)

    if highlights:
        global HCOLORS
        if HCOLORS is None:
            HCOLORS = [k for k in mcolors.TABLEAU_COLORS.keys() if "red" not in k]

        for i, (name, args) in enumerate(highlights):
            color_i = HCOLORS[i % len(HCOLORS)]
            highlight(ax, name, args, "black", color_i, color_i)

    if goal:
        name, args = goal
        lcolor = color1 = color2 = "red"
        highlight(ax, name, args, lcolor, color1, color2)


THEME = "dark"


def set_theme(theme) -> None:
    global THEME
    THEME = theme


def get_theme() -> str:
    return THEME


def draw(
    points: list[gm.Point],
    lines: list[gm.Line],
    circles: list[gm.Circle],
    segments: list[gm.Segment],
    goal: Any = None,
    highlights: list[tuple[str, list[gm.Point]]] = None,
    equals: list[tuple[Any, Any]] = None,
    block: bool = True,
    save_to: str = None,
    theme: str = "dark",
) -> None:
    """Draw everything on the same canvas."""
    plt.close()
    imsize = 512 / 100
    fig, ax = plt.subplots(figsize=(imsize, imsize), dpi=100)

    set_theme(theme)

    if get_theme() == "dark":
        ax.set_facecolor((0.0, 0.0, 0.0))
    else:
        ax.set_facecolor((1.0, 1.0, 1.0))

    _draw(ax, points, lines, circles, goal, equals, highlights)

    plt.axis("equal")
    fig.subplots_adjust(left=0, right=1, top=1, bottom=0, wspace=0, hspace=0)
    if points:
        xmin = min([p.num.x for p in points])
        xmax = max([p.num.x for p in points])
        ymin = min([p.num.y for p in points])
        ymax = max([p.num.y for p in points])
        plt.margins((xmax - xmin) * 0.1, (ymax - ymin) * 0.1)

    if save_to is not None:
        fig.savefig(save_to)

    plt.show(block=block)


def close_enough(a: float, b: float, tol: float = 1e-12) -> bool:
    return abs(a - b) < tol


def assert_close_enough(a: float, b: float, tol: float = 1e-12) -> None:
    assert close_enough(a, b, tol), f"|{a}-{b}| = {abs(a-b)} >= {tol}"


def ang_of(tail: Point, head: Point) -> float:
    vector = head - tail
    arctan = np.arctan2(vector.y, vector.x) % (2 * np.pi)
    return arctan


def ang_between(tail: Point, head1: Point, head2: Point) -> float:
    ang1 = ang_of(tail, head1)
    ang2 = ang_of(tail, head2)
    diff = ang1 - ang2
    # return diff % (2*np.pi)
    if diff > np.pi:
        return diff - 2 * np.pi
    if diff < -np.pi:
        return 2 * np.pi + diff
    return diff


def head_from(tail: Point, ang: float, length: float = 1) -> Point:
    vector = Point(np.cos(ang) * length, np.sin(ang) * length)
    return tail + vector


def random_points(n: int = 3) -> list[Point]:
    return [Point(unif(-1, 1), unif(-1, 1)) for _ in range(n)]


def random_rfss(*points: Point) -> list[Point]:
    """Random rotate-flip-scale-shift a point cloud."""
    # center point cloud.
    average = sum(points, Point(0.0, 0.0)) * (1.0 / len(points))
    points = [p - average for p in points]

    # rotate
    ang = unif(0.0, 2 * np.pi)
    sin, cos = np.sin(ang), np.cos(ang)
    # scale and shift
    scale = unif(0.5, 2.0)
    shift = Point(unif(-1, 1), unif(-1, 1))
    points = [p.rotate(sin, cos) * scale + shift for p in points]

    # randomly flip
    if np.random.rand() < 0.5:
        points = [p.flip() for p in points]

    return points


def reduce(
    objs: list[Union[Point, Line, Circle, HalfLine, HoleCircle]],
    existing_points: list[Point],
) -> list[Point]:
    """Reduce intersecting objects into one point of intersections."""
    if all(isinstance(o, Point) for o in objs):
        return objs

    elif len(objs) == 1:
        return objs[0].sample_within(existing_points)

    elif len(objs) == 2:
        a, b = objs
        result = a.intersect(b)
        if isinstance(result, Point):
            return [result]
        a, b = result
        a_close = any([a.close(x) for x in existing_points])
        if a_close:
            return [b]
        b_close = any([b.close(x) for x in existing_points])
        if b_close:
            return [a]
        return [np.random.choice([a, b])]

    else:
        raise ValueError(f"Cannot reduce {objs}")


def sketch(
    name: str, args: list[Union[Point, gm.Point]]
) -> list[Union[Point, Line, Circle, HalfLine, HoleCircle]]:
    fun = globals()["sketch_" + name]
    args = [p.num if isinstance(p, gm.Point) else p for p in args]
    out = fun(args)

    # out can be one or multiple {Point/Line/HalfLine}
    if isinstance(out, (tuple, list)):
        return list(out)
    return [out]


def sketch_on_opline(args: tuple[gm.Point, ...]) -> HalfLine:
    a, b = args
    return HalfLine(a, a + a - b)


def sketch_on_hline(args: tuple[gm.Point, ...]) -> HalfLine:
    a, b = args
    return HalfLine(a, b)


def sketch_ieq_triangle(args: tuple[gm.Point, ...]) -> tuple[Point, ...]:
    a = Point(0.0, 0.0)
    b = Point(1.0, 0.0)

    c, _ = Circle(a, p1=b).intersect(Circle(b, p1=a))
    return a, b, c


def sketch_incenter2(args: tuple[gm.Point, ...]) -> tuple[Point, ...]:
    a, b, c = args
    l1 = sketch_bisect([b, a, c])
    l2 = sketch_bisect([a, b, c])
    i = line_line_intersection(l1, l2)
    x = i.foot(Line(b, c))
    y = i.foot(Line(c, a))
    z = i.foot(Line(a, b))
    return x, y, z, i


def sketch_excenter2(args: tuple[gm.Point, ...]) -> tuple[Point, ...]:
    a, b, c = args
    l1 = sketch_bisect([b, a, c])
    l2 = sketch_exbisect([a, b, c])
    i = line_line_intersection(l1, l2)
    x = i.foot(Line(b, c))
    y = i.foot(Line(c, a))
    z = i.foot(Line(a, b))
    return x, y, z, i


def sketch_centroid(args: tuple[gm.Point, ...]) -> tuple[Point, ...]:
    a, b, c = args
    x = (b + c) * 0.5
    y = (c + a) * 0.5
    z = (a + b) * 0.5
    i = line_line_intersection(Line(a, x), Line(b, y))
    return x, y, z, i


def sketch_ninepoints(args: tuple[gm.Point, ...]) -> tuple[Point, ...]:
    a, b, c = args
    x = (b + c) * 0.5
    y = (c + a) * 0.5
    z = (a + b) * 0.5
    c = Circle(p1=x, p2=y, p3=z)
    return x, y, z, c.center


def sketch_2l1c(args: tuple[gm.Point, ...]) -> tuple[Point, ...]:
    """Sketch a circle touching two lines and another circle."""
    a, b, c, p = args
    bc, ac = Line(b, c), Line(a, c)
    circle = Circle(p, p1=a)

    d, d_ = line_circle_intersection(p.perpendicular_line(bc), circle)
    if bc.diff_side(d_, a):
        d = d_

    e, e_ = line_circle_intersection(p.perpendicular_line(ac), circle)
    if ac.diff_side(e_, b):
        e = e_

    df = d.perpendicular_line(Line(p, d))
    ef = e.perpendicular_line(Line(p, e))
    f = line_line_intersection(df, ef)

    g, g_ = line_circle_intersection(Line(c, f), circle)
    if bc.same_side(g_, a):
        g = g_

    b_ = c + (b - c) / b.distance(c)
    a_ = c + (a - c) / a.distance(c)
    m = (a_ + b_) * 0.5
    x = line_line_intersection(Line(c, m), Line(p, g))
    return x.foot(ac), x.foot(bc), g, x


def sketch_3peq(args: tuple[gm.Point, ...]) -> tuple[Point, ...]:
    a, b, c = args
    ab, _, ca = Line(a, b), Line(b, c), Line(c, a)

    z = b + (c - b) * np.random.uniform(-0.5, 1.5)

    z_ = z * 2 - c
    ca_parallel_line = z_.parallel_line(ca)
    x = line_line_intersection(ca_parallel_line, ab)
    y = z * 2 - x
    return x, y, z


def try_to_sketch_intersect(
    name1: str,
    args1: list[Union[gm.Point, Point]],
    name2: str,
    args2: list[Union[gm.Point, Point]],
    existing_points: list[Point],
) -> Optional[Point]:
    """Try to sketch an intersection between two objects."""
    obj1 = sketch(name1, args1)[0]
    obj2 = sketch(name2, args2)[0]

    if isinstance(obj1, Line) and isinstance(obj2, Line):
        fn = line_line_intersection
    elif isinstance(obj1, Circle) and isinstance(obj2, Circle):
        fn = circle_circle_intersection
    else:
        fn = line_circle_intersection
        if isinstance(obj2, Line) and isinstance(obj1, Circle):
            obj1, obj2 = obj2, obj1

    x = fn(obj1, obj2)

    if isinstance(x, Point):
        return x

    x1, x2 = x

    close1 = check_too_close([x1], existing_points)
    far1 = check_too_far([x1], existing_points)
    if not close1 and not far1:
        return x1
    close2 = check_too_close([x2], existing_points)
    far2 = check_too_far([x2], existing_points)
    if not close2 and not far2:
        return x2

    return None


def sketch_acircle(args: tuple[gm.Point, ...]) -> Circle:
    a, b, c, d, f = args
    de = sketch_aline([c, a, b, f, d])
    fe = sketch_aline([a, c, b, d, f])
    e = line_line_intersection(de, fe)
    return Circle(p1=d, p2=e, p3=f)


def sketch_aline(args: tuple[gm.Point, ...]) -> HalfLine:
    """Sketch the construction aline."""
    A, B, C, D, E = args
    ab = A - B
    cb = C - B
    de = D - E

    dab = A.distance(B)
    ang_ab = np.arctan2(ab.y / dab, ab.x / dab)

    dcb = C.distance(B)
    ang_bc = np.arctan2(cb.y / dcb, cb.x / dcb)

    dde = D.distance(E)
    ang_de = np.arctan2(de.y / dde, de.x / dde)

    ang_ex = ang_de + ang_bc - ang_ab
    X = E + Point(np.cos(ang_ex), np.sin(ang_ex))
    return HalfLine(E, X)


def sketch_amirror(args: tuple[gm.Point, ...]) -> HalfLine:
    """Sketch the angle mirror."""
    A, B, C = args
    ab = A - B
    cb = C - B

    dab = A.distance(B)
    ang_ab = np.arctan2(ab.y / dab, ab.x / dab)
    dcb = C.distance(B)
    ang_bc = np.arctan2(cb.y / dcb, cb.x / dcb)

    ang_bx = 2 * ang_bc - ang_ab
    X = B + Point(np.cos(ang_bx), np.sin(ang_bx))
    return HalfLine(B, X)


def sketch_bisect(args: tuple[gm.Point, ...]) -> Line:
    a, b, c = args
    ab = a.distance(b)
    bc = b.distance(c)
    x = b + (c - b) * (ab / bc)
    m = (a + x) * 0.5
    return Line(b, m)


def sketch_exbisect(args: tuple[gm.Point, ...]) -> Line:
    a, b, c = args
    return sketch_bisect(args).perpendicular_line(b)


def sketch_bline(args: tuple[gm.Point, ...]) -> Line:
    a, b = args
    m = (a + b) * 0.5
    return m.perpendicular_line(Line(a, b))


def sketch_dia(args: tuple[gm.Point, ...]) -> Circle:
    a, b = args
    return Circle((a + b) * 0.5, p1=a)


def sketch_tangent(args: tuple[gm.Point, ...]) -> tuple[Point, Point]:
    a, o, b = args
    dia = sketch_dia([a, o])
    return circle_circle_intersection(Circle(o, p1=b), dia)


def sketch_circle(args: tuple[gm.Point, ...]) -> Circle:
    a, b, c = args
    return Circle(center=a, radius=b.distance(c))


def sketch_cc_tangent(args: tuple[gm.Point, ...]) -> tuple[Point, ...]:
    """Sketch tangents to two circles."""
    o, a, w, b = args
    ra, rb = o.distance(a), w.distance(b)

    ow = Line(o, w)
    if close_enough(ra, rb):
        oo = ow.perpendicular_line(o)
        oa = Circle(o, ra)
        x, z = line_circle_intersection(oo, oa)
        y = x + w - o
        t = z + w - o
        return x, y, z, t

    swap = rb > ra
    if swap:
        o, a, w, b = w, b, o, a
        ra, rb = rb, ra

    oa = Circle(o, ra)
    q = o + (w - o) * ra / (ra - rb)

    x, z = circle_circle_intersection(sketch_dia([o, q]), oa)
    y = w.foot(Line(x, q))
    t = w.foot(Line(z, q))

    if swap:
        x, y, z, t = y, x, t, z

    return x, y, z, t


def sketch_hcircle(args: tuple[gm.Point, ...]) -> HoleCircle:
    a, b = args
    return HoleCircle(center=a, radius=a.distance(b), hole=b)


def sketch_e5128(args: tuple[gm.Point, ...]) -> tuple[Point, Point]:
    a, b, c, d = args

    g = (a + b) * 0.5
    de = Line(d, g)

    e, f = line_circle_intersection(de, Circle(c, p1=b))

    if e.distance(d) < f.distance(d):
        e = f
    return e, g


def sketch_eq_quadrangle(args: tuple[gm.Point, ...]) -> tuple[Point, ...]:
    """Sketch quadrangle with two equal opposite sides."""
    a = Point(0.0, 0.0)
    b = Point(1.0, 0.0)

    length = np.random.uniform(0.5, 2.0)
    ang = np.random.uniform(np.pi / 3, np.pi * 2 / 3)
    d = head_from(a, ang, length)

    ang = ang_of(b, d)
    ang = np.random.uniform(ang / 10, ang / 9)
    c = head_from(b, ang, length)
    a, b, c, d = random_rfss(a, b, c, d)
    return a, b, c, d


def sketch_eq_trapezoid(args: tuple[gm.Point, ...]) -> tuple[Point, ...]:
    a = Point(0.0, 0.0)
    b = Point(1.0, 0.0)
    lenght = unif(0.5, 2.0)
    height = unif(0.5, 2.0)
    c = Point(0.5 + lenght / 2.0, height)
    d = Point(0.5 - lenght / 2.0, height)

    a, b, c, d = random_rfss(a, b, c, d)
    return a, b, c, d


def sketch_eqangle2(args: tuple[gm.Point, ...]) -> Point:
    """Sketch the def eqangle2."""
    a, b, c = args

    ba = b.distance(a)
    bc = b.distance(c)
    lenght = ba * ba / bc

    if unif(0.0, 1.0) < 0.5:
        be = min(lenght, bc)
        be = unif(be * 0.1, be * 0.9)
    else:
        be = max(lenght, bc)
        be = unif(be * 1.1, be * 1.5)

    e = b + (c - b) * (be / bc)
    y = b + (a - b) * (be / lenght)
    return line_line_intersection(Line(c, y), Line(a, e))


def sketch_eqangle3(args: tuple[gm.Point, ...]) -> Circle:
    a, b, d, e, f = args
    de = d.distance(e)
    ef = e.distance(f)
    ab = b.distance(a)
    ang_ax = ang_of(a, b) + ang_between(e, d, f)
    x = head_from(a, ang_ax, length=de / ef * ab)
    return Circle(p1=a, p2=b, p3=x)


def sketch_eqdia_quadrangle(args: tuple[gm.Point, ...]) -> tuple[Point, ...]:
    """Sketch quadrangle with two equal diagonals."""
    m = unif(0.3, 0.7)
    n = unif(0.3, 0.7)
    a = Point(-m, 0.0)
    c = Point(1 - m, 0.0)
    b = Point(0.0, -n)
    d = Point(0.0, 1 - n)

    ang = unif(-0.25 * np.pi, 0.25 * np.pi)
    sin, cos = np.sin(ang), np.cos(ang)
    b = b.rotate(sin, cos)
    d = d.rotate(sin, cos)
    a, b, c, d = random_rfss(a, b, c, d)
    return a, b, c, d


def sketch_free(args: tuple[gm.Point, ...]) -> Point:
    return random_points(1)[0]


def sketch_isos(args: tuple[gm.Point, ...]) -> tuple[Point, ...]:
    base = unif(0.5, 1.5)
    height = unif(0.5, 1.5)

    b = Point(-base / 2, 0.0)
    c = Point(base / 2, 0.0)
    a = Point(0.0, height)
    a, b, c = random_rfss(a, b, c)
    return a, b, c


def sketch_line(args: tuple[gm.Point, ...]) -> Line:
    a, b = args
    return Line(a, b)


def sketch_cyclic(args: tuple[gm.Point, ...]) -> Circle:
    a, b, c = args
    return Circle(p1=a, p2=b, p3=c)


def sketch_hline(args: tuple[gm.Point, ...]) -> HalfLine:
    a, b = args
    return HalfLine(a, b)


def sketch_midp(args: tuple[gm.Point, ...]) -> Point:
    a, b = args
    return (a + b) * 0.5


def sketch_pentagon(args: tuple[gm.Point, ...]) -> tuple[Point, ...]:
    points = [Point(1.0, 0.0)]
    ang = 0.0

    for i in range(4):
        ang += (2 * np.pi - ang) / (5 - i) * unif(0.5, 1.5)
        point = Point(np.cos(ang), np.sin(ang))
        points.append(point)

    a, b, c, d, e = points
    a, b, c, d, e = random_rfss(a, b, c, d, e)
    return a, b, c, d, e


def sketch_pline(args: tuple[gm.Point, ...]) -> Line:
    a, b, c = args
    return a.parallel_line(Line(b, c))


def sketch_pmirror(args: tuple[gm.Point, ...]) -> Point:
    a, b = args
    return b * 2 - a


def sketch_quadrangle(args: tuple[gm.Point, ...]) -> tuple[Point, ...]:
    """Sketch a random quadrangle."""
    m = unif(0.3, 0.7)

    a = Point(-m, 0.0)
    c = Point(1 - m, 0.0)
    b = Point(0.0, -unif(0.25, 0.75))
    d = Point(0.0, unif(0.25, 0.75))

    ang = unif(-0.25 * np.pi, 0.25 * np.pi)
    sin, cos = np.sin(ang), np.cos(ang)
    b = b.rotate(sin, cos)
    d = d.rotate(sin, cos)
    a, b, c, d = random_rfss(a, b, c, d)
    return a, b, c, d


def sketch_r_trapezoid(args: tuple[gm.Point, ...]) -> tuple[Point, ...]:
    a = Point(0.0, 1.0)
    d = Point(0.0, 0.0)
    b = Point(unif(0.5, 1.5), 1.0)
    c = Point(unif(0.5, 1.5), 0.0)
    a, b, c, d = random_rfss(a, b, c, d)
    return a, b, c, d


def sketch_r_triangle(args: tuple[gm.Point, ...]) -> tuple[Point, ...]:
    a = Point(0.0, 0.0)
    b = Point(0.0, unif(0.5, 2.0))
    c = Point(unif(0.5, 2.0), 0.0)
    a, b, c = random_rfss(a, b, c)
    return a, b, c


def sketch_rectangle(args: tuple[gm.Point, ...]) -> tuple[Point, ...]:
    a = Point(0.0, 0.0)
    b = Point(0.0, 1.0)
    lenght = unif(0.5, 2.0)
    c = Point(lenght, 1.0)
    d = Point(lenght, 0.0)
    a, b, c, d = random_rfss(a, b, c, d)
    return a, b, c, d


def sketch_reflect(args: tuple[gm.Point, ...]) -> Point:
    a, b, c = args
    m = a.foot(Line(b, c))
    return m * 2 - a


def sketch_risos(args: tuple[gm.Point, ...]) -> tuple[Point, ...]:
    a = Point(0.0, 0.0)
    b = Point(0.0, 1.0)
    c = Point(1.0, 0.0)
    a, b, c = random_rfss(a, b, c)
    return a, b, c


def sketch_rotaten90(args: tuple[gm.Point, ...]) -> Point:
    a, b = args
    ang = -np.pi / 2
    return a + (b - a).rotate(np.sin(ang), np.cos(ang))


def sketch_rotatep90(args: tuple[gm.Point, ...]) -> Point:
    a, b = args
    ang = np.pi / 2
    return a + (b - a).rotate(np.sin(ang), np.cos(ang))


def sketch_s_angle(args: tuple[gm.Point, ...]) -> HalfLine:
    a, b, y = args
    ang = y / 180 * np.pi
    x = b + (a - b).rotatea(ang)
    return HalfLine(b, x)


def sketch_segment(args: tuple[gm.Point, ...]) -> tuple[Point, Point]:
    a, b = random_points(2)
    return a, b


def sketch_shift(args: tuple[gm.Point, ...]) -> Point:
    a, b, c = args
    return c + (b - a)


def sketch_square(args: tuple[gm.Point, ...]) -> tuple[Point, Point]:
    a, b = args
    c = b + (a - b).rotatea(-np.pi / 2)
    d = a + (b - a).rotatea(np.pi / 2)
    return c, d


def sketch_isquare(args: tuple[gm.Point, ...]) -> tuple[Point, ...]:
    a = Point(0.0, 0.0)
    b = Point(1.0, 0.0)
    c = Point(1.0, 1.0)
    d = Point(0.0, 1.0)
    a, b, c, d = random_rfss(a, b, c, d)
    return a, b, c, d


def sketch_tline(args: tuple[gm.Point, ...]) -> Line:
    a, b, c = args
    return a.perpendicular_line(Line(b, c))


def sketch_trapezoid(args: tuple[gm.Point, ...]) -> tuple[Point, ...]:
    d = Point(0.0, 0.0)
    c = Point(1.0, 0.0)

    base = unif(0.5, 2.0)
    height = unif(0.5, 2.0)
    a = Point(unif(0.2, 0.5), height)
    b = Point(a.x + base, height)
    a, b, c, d = random_rfss(a, b, c, d)
    return a, b, c, d


def sketch_triangle(args: tuple[gm.Point, ...]) -> tuple[Point, ...]:
    a = Point(0.0, 0.0)
    b = Point(1.0, 0.0)
    ac = unif(0.5, 2.0)
    ang = unif(0.2, 0.8) * np.pi
    c = head_from(a, ang, ac)
    return a, b, c


def sketch_triangle12(args: tuple[gm.Point, ...]) -> tuple[Point, ...]:
    b = Point(0.0, 0.0)
    c = Point(unif(1.5, 2.5), 0.0)
    a, _ = circle_circle_intersection(Circle(b, 1.0), Circle(c, 2.0))
    a, b, c = random_rfss(a, b, c)
    return a, b, c


def sketch_trisect(args: tuple[gm.Point, ...]) -> tuple[Point, Point]:
    """Sketch two trisectors of an angle."""
    a, b, c = args
    ang1 = ang_of(b, a)
    ang2 = ang_of(b, c)

    swap = 0
    if ang1 > ang2:
        ang1, ang2 = ang2, ang1
        swap += 1

    if ang2 - ang1 > np.pi:
        ang1, ang2 = ang2, ang1 + 2 * np.pi
        swap += 1

    angx = ang1 + (ang2 - ang1) / 3
    angy = ang2 - (ang2 - ang1) / 3

    x = b + Point(np.cos(angx), np.sin(angx))
    y = b + Point(np.cos(angy), np.sin(angy))

    ac = Line(a, c)
    x = line_line_intersection(Line(b, x), ac)
    y = line_line_intersection(Line(b, y), ac)

    if swap == 1:
        return y, x
    return x, y


def sketch_trisegment(args: tuple[gm.Point, ...]) -> tuple[Point, Point]:
    a, b = args
    x, y = a + (b - a) * (1.0 / 3), a + (b - a) * (2.0 / 3)
    return x, y



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Utilities for string manipulation in the DSL."""

MAP_SYMBOL = {
    "T": "perp",
    "P": "para",
    "D": "cong",
    "S": "simtri",
    "I": "circle",
    "M": "midp",
    "O": "cyclic",
    "C": "coll",
    "^": "eqangle",
    "/": "eqratio",
    "%": "eqratio",
    "=": "contri",
    "X": "collx",
    "A": "acompute",
    "R": "rcompute",
    "Q": "fixc",
    "E": "fixl",
    "V": "fixb",
    "H": "fixt",
    "Z": "fixp",
    "Y": "ind",
}


def map_symbol(c: str) -> str:
    return MAP_SYMBOL[c]


def map_symbol_inv(c: str) -> str:
    return {v: k for k, v in MAP_SYMBOL.items()}[c]


def _gcd(x: int, y: int) -> int:
    while y:
        x, y = y, x % y
    return x


def simplify(n: int, d: int) -> tuple[int, int]:
    g = _gcd(n, d)
    return (n // g, d // g)


def pretty2r(a: str, b: str, c: str, d: str) -> str:
    if b in (c, d):
        a, b = b, a

    if a == d:
        c, d = d, c

    return f"{a} {b} {c} {d}"


def pretty2a(a: str, b: str, c: str, d: str) -> str:
    if b in (c, d):
        a, b = b, a

    if a == d:
        c, d = d, c

    return f"{a} {b} {c} {d}"


def pretty_angle(a: str, b: str, c: str, d: str) -> str:
    if b in (c, d):
        a, b = b, a
    if a == d:
        c, d = d, c

    if a == c:
        return f"\u2220{b}{a}{d}"
    return f"\u2220({a}{b}-{c}{d})"


def pretty_nl(name: str, args: list[str]) -> str:
    """Natural lang formatting a predicate."""
    if name == "aconst":
        a, b, c, d, y = args
        return f"{pretty_angle(a, b, c, d)} = {y}"
    if name == "rconst":
        a, b, c, d, y = args
        return f"{a}{b}:{c}{d} = {y}"
    if name == "acompute":
        a, b, c, d = args
        return f"{pretty_angle(a, b, c, d)}"
    if name in ["coll", "C"]:
        return "" + ",".join(args) + " are collinear"
    if name == "collx":
        return "" + ",".join(list(set(args))) + " are collinear"
    if name in ["cyclic", "O"]:
        return "" + ",".join(args) + " are concyclic"
    if name in ["midp", "midpoint", "M"]:
        x, a, b = args
        return f"{x} is midpoint of {a}{b}"
    if name in ["eqangle", "eqangle6", "^"]:
        a, b, c, d, e, f, g, h = args
        return f"{pretty_angle(a, b, c, d)} = {pretty_angle(e, f, g, h)}"
    if name in ["eqratio", "eqratio6", "/"]:
        return "{}{}:{}{} = {}{}:{}{}".format(*args)
    if name == "eqratio3":
        a, b, c, d, o, o = args
        return f"S {o} {a} {b} {o} {c} {d}"
    if name in ["cong", "D"]:
        a, b, c, d = args
        return f"{a}{b} = {c}{d}"
    if name in ["perp", "T"]:
        if len(args) == 2:  # this is algebraic derivation.
            ab, cd = args  # ab = 'd( ... )'
            return f"{ab} \u27c2 {cd}"
        a, b, c, d = args
        return f"{a}{b} \u27c2 {c}{d}"
    if name in ["para", "P"]:
        if len(args) == 2:  # this is algebraic derivation.
            ab, cd = args  # ab = 'd( ... )'
            return f"{ab} \u2225 {cd}"
        a, b, c, d = args
        return f"{a}{b} \u2225 {c}{d}"
    if name in ["simtri2", "simtri", "simtri*"]:
        a, b, c, x, y, z = args
        return f"\u0394{a}{b}{c} is similar to \u0394{x}{y}{z}"
    if name in ["contri2", "contri", "contri*"]:
        a, b, c, x, y, z = args
        return f"\u0394{a}{b}{c} is congruent to \u0394{x}{y}{z}"
    if name in ["circle", "I"]:
        o, a, b, c = args
        return f"{o} is the circumcenter of \\Delta {a}{b}{c}"
    if name == "foot":
        a, b, c, d = args
        return f"{a} is the foot of {b} on {c}{d}"


def pretty(txt: str) -> str:
    """Pretty formating a predicate string."""
    if isinstance(txt, str):
        txt = txt.split(" ")
    name, *args = txt
    if name == "ind":
        return "Y " + " ".join(args)
    if name in ["fixc", "fixl", "fixb", "fixt", "fixp"]:
        return map_symbol_inv(name) + " " + " ".join(args)
    if name == "acompute":
        a, b, c, d = args
        return "A " + " ".join(args)
    if name == "rcompute":
        a, b, c, d = args
        return "R " + " ".join(args)
    if name == "aconst":
        a, b, c, d, y = args
        return f"^ {pretty2a(a, b, c, d)} {y}"
    if name == "rconst":
        a, b, c, d, y = args
        return f"/ {pretty2r(a, b, c, d)} {y}"
    if name == "coll":
        return "C " + " ".join(args)
    if name == "collx":
        return "X " + " ".join(args)
    if name == "cyclic":
        return "O " + " ".join(args)
    if name in ["midp", "midpoint"]:
        x, a, b = args
        return f"M {x} {a} {b}"
    if name == "eqangle":
        a, b, c, d, e, f, g, h = args
        return f"^ {pretty2a(a, b, c, d)} {pretty2a(e, f, g, h)}"
    if name == "eqratio":
        a, b, c, d, e, f, g, h = args
        return f"/ {pretty2r(a, b, c, d)} {pretty2r(e, f, g, h)}"
    if name == "eqratio3":
        a, b, c, d, o, o = args
        return f"S {o} {a} {b} {o} {c} {d}"
    if name == "cong":
        a, b, c, d = args
        return f"D {a} {b} {c} {d}"
    if name == "perp":
        if len(args) == 2:  # this is algebraic derivation.
            ab, cd = args  # ab = 'd( ... )'
            return f"T {ab} {cd}"
        a, b, c, d = args
        return f"T {a} {b} {c} {d}"
    if name == "para":
        if len(args) == 2:  # this is algebraic derivation.
            ab, cd = args  # ab = 'd( ... )'
            return f"P {ab} {cd}"
        a, b, c, d = args
        return f"P {a} {b} {c} {d}"
    if name in ["simtri2", "simtri", "simtri*"]:
        a, b, c, x, y, z = args
        return f"S {a} {b} {c} {x} {y} {z}"
    if name in ["contri2", "contri", "contri*"]:
        a, b, c, x, y, z = args
        return f"= {a} {b} {c} {x} {y} {z}"
    if name == "circle":
        o, a, b, c = args
        return f"I {o} {a} {b} {c}"
    if name == "foot":
        a, b, c, d = args
        return f"F {a} {b} {c} {d}"
    return " ".join(txt)



import newclid


from pathlib import Path


def default_configs_path() -> Path:
    return Path(newclid.__file__).parent.joinpath("default_configs")


def default_defs_path() -> Path:
    return default_configs_path().joinpath("defs.txt")


def default_rules_path() -> Path:
    return default_configs_path().joinpath("rules.txt")
